"Internal support for the text toolkit"

# Copyright (c) 2010-2021 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import absolute_import, print_function

import sys
import re

import _chemfp
from . import io
from . import base_toolkit

from ._compat import enumerate, next, unicode_type, bytes_type

BLAME_NAME = "The text toolkit"

def tag_argument_error(tag):
    if b"\n" in tag:
        raise ValueError("tag must not contain a newline")
    if b"<" in tag:
        raise ValueError("tag must not contain a '<'")
    if b">" in tag:
        raise ValueError("tag must not contain a '>'")
    if len(tag) > 200:
        raise ValueError("tag may not be longer than 200 characters")

def _fancy_repr(record):
    if record is not None:
        if len(record) > 60:
            record_s = repr(record[:57])
            record_s = record_s[:-1] + " ..." + record_s[-1]
        else:
            record_s = repr(record)
    else:
        record_s = "None"
    
    return record_s

class TextRecord(object):
    """Base class for the text_toolkit 'molecules', which work with the records as text.

    The :mod:`chemfp.text_toolkit` implements the toolkit API, but
    it doesn't know chemistry. Instead of returning real molecule objects, with
    atoms and bonds, it returns TextRecord subclass instances that hold the
    record as a text string.

    As an implementation detail (which means its subject to change) there is
    a subclass for each of the support formats.
    
    * :class:`SDFRecord` - holds "sdf" records
    * :class:`SmiRecord` - holds "smi" records (the full line from a "smi" SMILES file)
    * :class:`CanRecord` - holds "can" records (the full line from a "can" SMILES file)
    * :class:`UsmRecord` - holds "usm" records (the full line from a "usm" SMILES file)
    * :class:`SmiStringRecord` - holds "smistring" records (only the "smistring" SMILES string; no id)
    * :class:`CanStringRecord` - holds "canstring" records (only the "canstring" SMILES string; no id)
    * :class:`UsmStringRecord` - holds "usmstring" records (only the "usmstring" SMILES string; no id)

    All of the classes have the following attributes:
    .. py:attribute:: id

       The record identifier as a Unicode string, or None if there is no identifier

    .. py:attribute:: id_bytes

       The record identifier as a byte string, or None if there is no identifier

    .. py:attribute:: record

       The record, as a string. For the smistring, canstring, and usmstring
       formats, this is only the SMILES string.

    .. py:attribute:: record_format

       One of "sdf", "smi", "can", "usm", "smistring", "canstring", or "usmstring".

    The SMILES classes have an attribute:

    .. py:attribute:: smiles

       The SMILES string component of the record.

    """

    __init__ = NotImplemented  # Must call a child class
    
    def add_tag(self, tag, value):
        """Add an SD tag value to the TextRecord

        This methods does nothing if the record is not an "sdf" record.

        :param tag: the SD tag name
        :type tag: string
        :param value: the text for the tag
        :type value: string
        :returns: None
        """
        pass
    
    def get_tag(self, tag):
        """Get the named SD tag value, or None if it doesn't exist or is not an "sdf" record.

        :param tag: the SD tag name
        :type tag: byte or Unicode string
        :returns: a Unicode string, or None
        """
        return None
    
    def get_tag_as_bytes(self, tag):
        """Get the named SD tag value, or None if it doesn't exist or is not an "sdf" record.

        :param tag: the SD tag name
        :type tag: byte string
        :returns: a byte string, or None
        """
        return None
    
    def get_tag_pairs(self):
        """Get a list of all SD tag (name, value) pairs for the TextRecord using Unicode strings

        This function returns an empty list if the record is not an "sdf" record.

        :returns: a list of (Unicode string name, Unicode string value) pairs
        """
        return []
    
    def get_tag_pairs_as_bytes(self):
        """Get a list of all SD tag (name, value) pairs for the TextRecord using byte strings

        This function returns an empty list if the record is not an "sdf" record.

        :returns: a list of (byte string name, byte string value) pairs
        """
        return []

    def copy(self):
        """Return a new record which is a copy of the given record"""
        raise NotImplementedError


class SDFRecord(TextRecord):
    "Holds an SDF record. See :class:`chemfp._text_toolkit.TextRecord` for API details"
    record_format = "sdf"
    __slots__ = ("_id", "_id_bytes", "record", "encoding", "encoding_errors")
    def __init__(self, id_bytes, record, encoding, encoding_errors):
        self._id_bytes = id_bytes
        try:
            self._id = id_bytes.decode(encoding, encoding_errors)
        except UnicodeDecodeError:
            self._id = None
        self.record = record
        self.encoding = encoding
        self.encoding_errors = encoding_errors


    def copy(self):
        return SDFRecord(self.id_bytes, self.record, self.encoding, self.encoding_errors)

    def __repr__(self):
        return "SDFRecord(id_bytes=%r(id=%r), record=%s, encoding=%r, encoding_errors=%r)" % (
            self.id_bytes, self.id, _fancy_repr(self.record), self.encoding, self.encoding_errors)

    @property
    def id(self):
        return self._id
    @id.setter
    def id(self, id):
        self._id_bytes = id.encode(self.encoding, self.encoding_errors)
        self._id = id
        
    @property
    def id_bytes(self):
        return self._id_bytes
    @id_bytes.setter
    def id_bytes(self, id_bytes):
        self._id = id_bytes.decode(self.encoding, self.encoding_errors)
        self._id_bytes = id_bytes
    
    def add_tag(self, tag, value):
        if isinstance(tag, unicode_type):
            tag = tag.encode(self.encoding, self.encoding_errors)
        if isinstance(value, unicode_type):
            value = value.encode(self.encoding, self.encoding_errors)
                             
        self.record = add_sdf_record_tag(self.record, tag, value)
    
    def get_tag(self, tag):
        return get_sdf_record_tag(self.record, tag, self.encoding, self.encoding_errors)
    
    def get_tag_as_bytes(self, tag):
        return get_sdf_record_tag_as_bytes(self.record, tag)
        
    def get_tag_pairs(self, encoding=None, encoding_errors=None):
        if encoding is None:
            encoding = self.encoding
        if encoding_errors is None:
            encoding_errors = self.encoding_errors
        return [(k.decode(encoding, encoding_errors), v.decode(encoding, encoding_errors))
                     for (k, v) in get_sdf_tag_pairs(self.record)]
    
    def get_tag_pairs_as_bytes(self):
        return get_sdf_tag_pairs(self.record)
    

    copy.__doc__ = TextRecord.copy.__doc__
    add_tag.__doc__ = TextRecord.add_tag.__doc__
    get_tag.__doc__ = TextRecord.get_tag.__doc__
    get_tag_pairs.__doc__ = TextRecord.get_tag_pairs.__doc__

class _SmiRecord(TextRecord):
    __slots__ = ("id", "record", "smiles", "encoding", "encoding_errors")
    def __init__(self, id, record, smiles, encoding, encoding_errors):
        self.id = id
        self.record = record
        self.smiles = smiles
        self.encoding = encoding
        self.encoding_errors = encoding_errors

    def copy(self):
        return self.__class__(self.id, self.record, self.smiles, self.encoding, self.encoding_errors)

    def __repr__(self):
        smiles_repr = _fancy_repr(self.smiles)
        record_repr = _fancy_repr(self.record)
        return "%s(id=%r, record=%s, smiles=%s, encoding=%r, encoding_errors=%r)" % (
            self.__class__.__name__, self.id, record_repr, smiles_repr, self.encoding, self.encoding_errors)
    
class SmiRecord(_SmiRecord):
    """Holds an "smi" record. See :class:`chemfp._text_toolkit.TextRecord` for API details"""
    record_format = "smi"

class CanRecord(_SmiRecord):
    """Holds an "can" record. See :class:`chemfp._text_toolkit.TextRecord` for API details"""
    record_format = "can"

class UsmRecord(_SmiRecord):
    """Holds an "usm" record. See :class:`chemfp._text_toolkit.TextRecord` for API details"""
    record_format = "usm"

class _SmiStringRecord(TextRecord):
    __slots__ = ("id", "record", "smiles")
    encoding = "utf8"
    encoding_errors = "strict"
    def __init__(self, id, record, smiles):
        self.id = id
        self.record = record
        self.smiles = smiles

    def copy(self):
        return self.__class__(self.id, self.record, self.smiles)

    def __repr__(self):
        record_repr = _fancy_repr(self.record)
        return "%s(id=%r, record=%s, smiles=%r)" % (
            self.__class__.__name__, self.id, record_repr, self.smiles)

class SmiStringRecord(_SmiStringRecord):
    """Holds an "smistring" record. See :class:`chemfp._text_toolkit.TextRecord` for API details"""
    record_format = "smistring"

class CanStringRecord(_SmiStringRecord):
    """Holds an "canstring" record. See :class:`chemfp._text_toolkit.TextRecord` for API details"""
    record_format = "canstring"

class UsmStringRecord(_SmiStringRecord):
    """Holds an "usmstring" record. See :class:`chemfp._text_toolkit.TextRecord` for API details"""
    record_format = "usmstring"
    


######### SMILES and InChI files


# The first field is a 'molecule', the second is the id.

def iter_first_two_delimited_fields_as_strings(infile, close, delimiter, has_header,
                                               error_handler, location, register_name,
                                               encoding, encoding_errors):

    it = iter_first_two_delimited_fields_as_bytes(infile, close, delimiter, has_header,
                                                  error_handler, location, register_name)
    def _iter_first_two_delimited_fields_as_strings(it, error_handler, encoding, encoding_errors):
        try:
            for smiles_bytes, id_bytes in it:
                if smiles_bytes is None:
                    smiles = None
                else:
                    try:
                        smiles = smiles_bytes.decode(encoding, encoding_errors)
                    except UnicodeDecodeError as err:
                        error_handler.error("Cannot decode the SMILES %r: %s" % (smiles_bytes, err),
                                            location)
                        continue
                if id_bytes is None:
                    id = None
                else:
                    try:
                        id = id_bytes.decode(encoding, encoding_errors)
                    except UnicodeDecodeError as err:
                        error_handler.error("Cannot decode the identifier %r: %s" % (id_bytes, err),
                                            location)
                        continue
                yield smiles, id
        finally:
            it.close()

    return _iter_first_two_delimited_fields_as_strings(it, error_handler, encoding, encoding_errors)
    

def iter_first_two_delimited_fields_as_bytes(infile, close, delimiter, has_header,
                                             error_handler, location, register_name):
    initial_lineno = 0
    initial_offset = 0
    if has_header:
        header = next(infile, None)
        if header is not None:
            initial_lineno = 1
            initial_offset += len(header)
    
    it = _iter_first_two_delimited_fields(infile, close, delimiter,
                                          initial_lineno, initial_offset,
                                          error_handler, location, register_name)
    next(it)
    return it

def _iter_first_two_delimited_fields(infile, close, delimiter,
                                     initial_lineno, initial_offset,
                                     error_handler, location, register_name):
    start_offset = end_offset = initial_offset
    lineno = initial_lineno
    terms = []
    raw_line = None
    def get_lineno():
        return lineno
    def get_recno():
        return lineno
    def get_offsets():
        return start_offset, end_offset
    def get_record():
        return raw_line
    
    location.register(get_lineno=get_lineno,
                      get_recno=get_recno,
                      get_offsets=get_offsets,
                      get_record=get_record)

    # Handle the delimiter code here, rather than though a special method.
    delimiter = delimiter.name
    if delimiter == u"\t":
        delimiter_bytes = b"\t"
    elif delimiter == u" ":
        delimiter_bytes = b" "
    elif len(delimiter) == 1:
        if hasattr(delimiter, "encode"):
            delimiter_bytes = delimiter.encode("ascii")
        else:
            delimiter_bytes = delimiter
    
    try:
        yield "Ready for processing!"

        if delimiter == "to-eol":
            for lineno, raw_line in enumerate(infile, lineno+1):
                start_offset = end_offset
                end_offset = start_offset + len(raw_line)
                line = raw_line.rstrip(b"\r\n")
                terms = line.split(None, 1)
                if not terms:
                    error_handler.error("Empty line in %s file" % (register_name,), location)
                    continue
                if len(terms) == 1:
                    # Missing id. What should we do?
                    yield terms[0], None
                else:
                    # By construction, terms[1] is not blank
                    yield terms[0], terms[1]

        elif delimiter == "whitespace":
            for lineno, raw_line in enumerate(infile, lineno+1):
                start_offset = end_offset
                end_offset = start_offset + len(raw_line)
                terms = raw_line.split()
                if not terms:
                    error_handler.error("Empty line in %s file" % (register_name,), location)
                    continue
                if len(terms) == 1:
                    # Missing id. What should we do?
                    yield terms[0], None
                else:
                    # By construction, terms[1] is not blank
                    yield terms[0], terms[1]

        elif len(delimiter) == 1:
            for lineno, raw_line in enumerate(infile, lineno+1):
                start_offset = end_offset
                end_offset = start_offset + len(raw_line)
                line = raw_line.rstrip(b"\r\n")
                if not line:
                    error_handler.error("Empty line in %s file" % (register_name,), location)
                    continue
                terms = line.split(delimiter_bytes, 2)
                if len(terms) == 1:
                    # Missing id. What should we do?
                    yield terms[0], None
                else:
                    # It's possible for this id to be ""
                    yield terms[0], terms[1]
        else:
            raise AssertionError("Unknown delimiter name %r" % (delimiter,))

    finally:
        # Save first, then close. This is the most likely to reduce memory cycles.
        location.save(lineno=get_lineno(),
                      recno=get_recno(),
                      offsets=(end_offset, end_offset),
                      record=get_record())
        if close is not None:
            try:
                close()
            finally:
                close = None

def iter_line_and_first_two_delimited_fields_as_strings(
        infile, close, delimiter, has_header,
        error_handler, location, register_name,
        encoding, encoding_errors):
    it = iter_line_and_first_two_delimited_fields(infile, close, delimiter, has_header,
                                                  error_handler, location, register_name)
    def iter_line_and_first_two_delimited_fields_as_strings(it, error_handler, encoding, encoding_errors):
        try:
            for line_bytes, smiles_bytes, id_bytes in it:
                if smiles_bytes is None:
                    smiles = None
                else:
                    try:
                        smiles = smiles_bytes.decode(encoding, encoding_errors)
                    except UnicodeDecodeError as err:
                        error_handler.error("Cannot decode the SMILES %r: %s" % (smiles_bytes, err),
                                            location)

                if id_bytes is None:
                    id = None
                else:
                    try:
                        id = id_bytes.decode(encoding, encoding_errors)
                    except UnicodeDecodeError as err:
                        error_handler.error("Cannot decode the identifier %r: %s" % (id_bytes, err),
                                            location)

                yield line_bytes, smiles, id
        finally:
            it.close()
    
    return iter_line_and_first_two_delimited_fields_as_strings(it, error_handler, encoding, encoding_errors)

def iter_line_and_first_two_delimited_fields(infile, close, delimiter, has_header,
                                             error_handler, location, register_name):
    infile = iter(infile)
    initial_lineno = 0
    initial_offset = 0
    if has_header:
        header = next(infile, None)
        if header is not None:
            initial_lineno = 1
            initial_offset += len(header)
    
    
    it = _iter_line_and_first_two_delimited_fields(infile, close, delimiter,
                                                   initial_lineno, initial_offset,
                                                   error_handler, location, register_name)
    next(it)
    return it

def _iter_line_and_first_two_delimited_fields(infile, close, delimiter,
                                              initial_lineno, initial_offset,
                                              error_handler, location, register_name):
    start_offset = end_offset = initial_offset
    lineno = initial_lineno
    terms = []
    raw_line = None
    def get_lineno():
        return lineno
    def get_recno():
        return lineno
    def get_offsets():
        return start_offset, end_offset
    def get_record():
        return raw_line
    
    location.register(get_lineno=get_lineno,
                      get_recno=get_recno,
                      get_offsets=get_offsets,
                      get_record=get_record)

    # Handle the delimiter code here, rather than though a special method.
    delimiter = delimiter.name
    if delimiter == u"\t":
        delimiter_bytes = b"\t"
    elif delimiter == u" ":
        delimiter_bytes = b" "
    elif len(delimiter) == 1:
        if hasattr(delimiter, "encode"):
            delimiter_bytes = delimiter.encode("ascii")
        else:
            delimiter_bytes = delimiter
    
    try:
        yield "Ready for processing!"

        if delimiter == "to-eol":
            for lineno, raw_line in enumerate(infile, lineno+1):
                start_offset = end_offset
                end_offset = start_offset + len(raw_line)
                line = raw_line.rstrip(b"\r\n")
                terms = line.split(None, 1)
                if not terms:
                    error_handler.error("Empty line in %s file" % (register_name,), location)
                    continue
                if len(terms) == 1:
                    # Missing id. What should we do?
                    yield raw_line, terms[0], None
                else:
                    # By construction, terms[1] is not blank
                    yield raw_line, terms[0], terms[1]

        elif delimiter == "whitespace":
            for lineno, raw_line in enumerate(infile, lineno+1):
                start_offset = end_offset
                end_offset = start_offset + len(raw_line)
                terms = raw_line.split()
                if not terms:
                    error_handler.error("Empty line in %s file" % (register_name,), location)
                    continue
                if len(terms) == 1:
                    # Missing id. What should we do?
                    yield raw_line, terms[0], None
                else:
                    # By construction, terms[1] is not blank
                    yield raw_line, terms[0], terms[1]

        elif len(delimiter) == 1:
            for lineno, raw_line in enumerate(infile, lineno+1):
                start_offset = end_offset
                end_offset = start_offset + len(raw_line)
                line = raw_line.rstrip(b"\r\n")
                if not line:
                    error_handler.error("Empty line in %s file" % (register_name,), location)
                    continue
                terms = line.split(delimiter_bytes, 2)
                if len(terms) == 1:
                    # Missing id. What should we do?
                    yield raw_line, terms[0], None
                else:
                    # It's possible for this id to be ""
                    yield raw_line, terms[0], terms[1]
        else:
            raise AssertionError("Unknown delimiter name %r" % (delimiter,))

    finally:
        # Save first, then close. This is the most likely to reduce memory cycles.
        location.save(lineno=get_lineno(),
                      recno=get_recno(),
                      offsets=(end_offset, end_offset),
                      record=get_record())
        if close is not None:
            try:
                close()
            finally:
                close = None


def open_two_delimited_fields_writer_gen(outfile, close, delimiter, error_handler, location,
                                         format_name, valid_formats, encoding, encoding_errors):
    it = _open_two_delimited_fields_writer_gen(outfile, close, delimiter, error_handler, location,
                                               format_name, valid_formats, encoding, encoding_errors)
    next(it)
    return it

def _open_two_delimited_fields_writer_gen(outfile, close, delimiter, error_handler, location,
                                          format_name, valid_formats, encoding, encoding_errors):
    recno = output_recno = 0
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    
    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno,
                      get_lineno=get_output_recno)

    two_arg_format = delimiter.get_format_string()
    
    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:

                # NOTE: the 'molecule' records here are MoleculeRecord instances.
                # This can't be used for arbitrary two-element tuples.
                if command == "write_molecule":
                    id, rec = args
                    if id is None:
                        command = "write_molecules"
                        args = [rec]
                    else:
                        command = "write_ids_and_molecules"
                        args = [(id, rec)]

                if command == "write_molecules":
                    for rec in args:
                        recno += 1  # 'recno' is the number of *input* records
                        if rec.record_format not in valid_formats:
                            error_handler.error(
                                "Cannot write a record in %r format to %s file"
                                % (rec.record_format, format_name), location)
                            continue

                        smiles = rec.smiles
                        if smiles is None:
                            error_handler.error("Missing SMILES", location)
                            continue
                        if "\n" in smiles:
                            error_handler.error("SMILES must not contain a newline", location)
                            continue

                        id = rec.id
                        if id is None:
                            id = ""
                        elif "\n" in id:
                            error_handler.error("Id must not contain a newline", location)
                            continue

                        if not outfile.write_string(two_arg_format % (smiles, id), error_handler, location):
                            continue
                        output_recno += 1

                elif command == "write_ids_and_molecules":
                    for id, rec in args:
                        recno += 1  # 'recno' is the number of *input* records
                        if rec.record_format not in valid_formats:
                            error_handler.error(
                                "Cannot write a record in %r format to %s file"
                                % (rec.record_format, format_name), location)
                            continue

                        smiles = rec.smiles
                        if smiles is None:
                            error_handler.error("Missing SMILES", location)
                            continue
                        if "\n" in smiles:
                            error_handler.error("SMILES must not contain a newline", location)
                            continue

                        if id is None:
                            id = rec.id
                            if id is None:
                                id = ""

                        if "\n" in id:
                            error_handler.error("Id must not contain a newline", location)
                            continue
                        if not outfile.write_string(two_arg_format % (smiles, id), error_handler, location):
                            continue
                        output_recno += 1

                else:
                    raise AssertionError("Unknown command: %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
    finally:
        location.save(recno = get_recno(),
                      output_recno = get_output_recno(),
                      lineno = get_output_recno())
                      
        if close is not None:
            try:
                close()
            finally:
                close = None
        

def _create_delimited_smi_string(format_config, mol, id, writer_args, errors):
    delimiter = base_toolkit.get_delimiter(format_config, writer_args)

    smiles = mol.smiles
    if smiles is None:
        error_handler = io.get_parse_error_handler(errors)
        error_handler.error("Missing SMILES")
        return None

    if "\n" in smiles:
        error_handler = io.get_parse_error_handler(errors)
        error_handler.error("SMILES must not contain a newline")
        return None

    if id is None:
        id = mol.id
        if id is None:
            id = ""

    if "\n" in id:
        error_handler = io.get_parse_error_handler(errors)
        error_handler.error("Id must not contain a newline")
        return None

    return delimiter.format_two_terms(smiles, id)


#########  SDF reader

def _check_block_size(block_size):
    if block_size <= 0:
        raise ValueError("block_size must be positive")

def get_id_tag_bytes(id_tag, encoding, encoding_errors):
    if id_tag is None:
        return None
    if isinstance(id_tag, unicode_type):
        id_tag_bytes = id_tag.encode(encoding, encoding_errors)
    elif isinstance(id_tag, bytes_type):
        id_tag_bytes = id_tag
    else:
        raise ValueError("id_tag must be a string, bytes, or None")

    if len(id_tag_bytes) > 200:
        raise ValueError("id_tag is too long (max 200 bytes)")
    if b"\n" in id_tag_bytes:
        raise ValueError("id_tag must not contain a newline")
    if b"\0" in id_tag_bytes:
        raise ValueError("id_tag must not contain a NUL character")
    return id_tag_bytes

def get_value_tag_bytes(value_tag, encoding, encoding_errors):
    if value_tag is None:
        return None
    if isinstance(value_tag, unicode_type):
        value_tag_bytes = value_tag.encode(encoding, encoding_errors)
    elif isinstance(value_tag, bytes_type):
        value_tag_bytes = value_tag
    else:
        raise ValueError("value_tag must be a string, bytes, or None")
    
    if len(value_tag_bytes) > 200:
        raise ValueError("value_tag is too long (max 200 bytes)")
    if b"\n" in value_tag_bytes:
        raise ValueError("value_tag must not contain a newline")
    if b"\0" in value_tag_bytes:
        raise ValueError("value_tag must not contain a NUL character")
    return value_tag_bytes


def iter_sdf_records(infile, close, error_handler, location, block_size):
    it = _iter_sdf_records(infile, close, error_handler, location, block_size)
    next(it)
    return it

def _iter_sdf_records(infile, close, error_handler, location, block_size):
    record = None
    recno = 0
    lineno = 1
    start_offset = end_offset = offset = 0

    # Register how to get ahold of the different parts
    def get_lineno():
        if recno == 0:
            return 0
        return lineno

    def get_recno():
        return recno

    def get_offsets():
        return start_offset, end_offset

    def get_record():
        return record

    location.register(get_lineno=get_lineno,
                      get_recno=get_recno,
                      get_offsets=get_offsets,
                      get_record=get_record)
    try:
        yield "Ready for processing!"
        block = b""
        while 1:

            # Read the next block.
            next_block = infile.read(block_size)

            if not next_block:
                # At the end of the file. Check for extra data.
                if not block:
                    return
                recno += 1
                record = block[offset:]
                error_handler.error("Incomplete sdf record or extra trailing data", location)
                return

            # Merge the current and next blocks
            block = block + next_block
            offset = 0

            while 1:
                # Figure out the contents and end of the current record.
                err, new_offset, id, num_newlines = _chemfp.find_sdf_record_one_field(block, offset, None)
                # ignore the id from the title line

                if err == 1:
                    # Found a record
                    record = block[offset:new_offset]
                    #assert record.count("\n") == num_newlines, (record.count("\n"), num_newlines)

                    # Provide location information then yield the id and record
                    recno += 1
                    yield record

                    # Update the offset information
                    start_offset = end_offset
                    end_offset += (new_offset-offset)

                    # Keep track of location information
                    lineno += num_newlines
                    offset = new_offset

                elif err == 2:
                    # No more content in this block. Reset, and fetch more data.
                    block = b""
                    break

                elif err == 3:
                    # Incomplete record. Need more data. Reset the block so it
                    # only contains the trailing data, and fetch more data.
                    block = block[offset:]
                    break

                elif err == 4:
                    # Parse error.
                    orig_lineno = lineno

                    lineno += num_newlines
                    recno += 1
                    start_offset = end_offset

                    # See if I can finish off this record
                    to_read = max(327680, 5*block_size) - len(block)
                    if to_read > 0:
                        block += infile.read(to_read)

                    i = block.find(b"\n$$$$\n", offset)
                    if i == -1:
                        end_offset += (new_offset-offset)
                        record = block[offset:]
                        error_handler.error("The text toolkit cannot parse the sdf record. It is "
                                            "not in the right format, and unable to continue",
                                            location)
                        return

                    # It looks like I can recover!
                    record = block[offset:i+6]
                    offset = i+6

                    # Update the location information and generate the error message
                    end_offset = start_offset + len(record)

                    error_handler.error("The text toolkit cannot parse the sdf record. It is "
                                        "not in the right format",
                                        location)

                    # Update to past the end of the bad record
                    lineno = orig_lineno + record.count(b"\n")

                    # Reset so a future get_record() doesn't return wrong information
                    record = None

                    # Continue processing the current block
                    continue


                elif err == 5:
                    # Found a record, but without an id. ... Shouldn't happen for this case?
                    record = block[offset:new_offset]
                    #assert record.count(b"\n") == num_newlines, (record.count(b"\n"), num_newlines)

                    start_offset = end_offset
                    end_offset += (new_offset-offset)

                    recno += 1
                    yield record

                    # Keep track of where we are
                    recno += 1
                    lineno += num_newlines
                    offset = new_offset

                elif err == 6:
                    raise AssertionError("Argument error but I can't figure out what happened")

                else:
                    raise AssertionError("Unknown error code %r" % (err,))
    finally:
        # I seem to be confused about how 'lineno' works
        final_lineno = get_lineno()
        if final_lineno > 0:
            final_lineno -= 1
        location.save(lineno = final_lineno,
                      recno = get_recno(),
                      offsets = get_offsets(),
                      record = get_record())
        if close is not None:
            try:
                close()
            finally:
                close = None
    


def iter_sdf_ids_and_records(infile, close, id_tag_bytes, error_handler, location,
                             encoding, encoding_errors, block_size=327680):
    assert not isinstance(id_tag_bytes, unicode_type), "must be bytes"
    it = _iter_sdf_id_bytes_and_records(infile, close, id_tag_bytes, error_handler, location,
                                        encoding, encoding_errors, block_size)
    next(it)
    def iter_sdf_ids_and_records():
        try:
            for id_bytes, record in it:
                if id_bytes is not None:
                    try:
                        id = id_bytes.decode(encoding, encoding_errors)
                    except UnicodeDecodeError as err:
                        # encoding_errors is 'strict', so report that as a true error
                        error_handler.error("Cannot decode the identifier %r: %s" % (id_bytes, err),
                                            location)
                        # I can recover from it by doing nothing
                        continue
                    yield id, record
                else:
                    yield None, record
        finally:
            it.close()
    
    return iter_sdf_ids_and_records()

def iter_sdf_id_bytes_and_records(infile, close, id_tag_bytes, error_handler, location,
                                  encoding, encoding_errors, block_size=327680):
    it = _iter_sdf_id_bytes_and_records(infile, close, id_tag_bytes, error_handler, location,
                                        encoding, encoding_errors, block_size)
    next(it)
    return it

def _iter_sdf_id_bytes_and_records(infile, close, id_tag_bytes, error_handler, location,
                                   encoding, encoding_errors, block_size):
    record = None
    recno = 0
    lineno = 1
    start_offset = end_offset = offset = 0

    # Register how to get ahold of the different parts
    def get_lineno():
        if recno == 0:
            return 0
        return lineno

    def get_recno():
        return recno

    def get_offsets():
        return start_offset, end_offset

    def get_record():
        return record

    location.register(get_lineno=get_lineno,
                      get_recno=get_recno,
                      get_offsets=get_offsets,
                      get_record=get_record)
        
    try:
        yield "Ready for processing!"
        
        block = b""

        while 1:
            # Read the next block.
            next_block = infile.read(block_size)
            
            if not next_block:
                # At the end of the file. Check for extra data.
                if not block:
                    return
                recno += 1
                record = block[offset:]
                error_handler.error("Incomplete sdf record or extra trailing data", location)
                return


            # Merge the current and next blocks
            block = block + next_block
            offset = 0

            while 1:
                # Figure out the contents and end of the current record.
                err, new_offset, id_bytes, num_newlines = _chemfp.find_sdf_record_one_field(
                    block, offset, id_tag_bytes)
                #

                if err == 1:
                    # Found a record
                    record = block[offset:new_offset]
                    #assert record.count(b"\n") == num_newlines, (record.count(b"\n"), num_newlines)

                    # Provide location information then yield the id and record
                    recno += 1
                    end_offset += (new_offset-offset)
                    yield (id_bytes, record)
                    
                    # Update the offset information
                    start_offset = end_offset

                    # Keep track of location information
                    lineno += num_newlines
                    offset = new_offset

                elif err == 2:
                    # No more content in this block. Reset, and fetch more data.
                    block = b""
                    break

                elif err == 3:
                    # Incomplete record. Need more data. Reset the block so it
                    # only contains the trailing data, and fetch more data.
                    block = block[offset:]
                    break

                elif err == 4:
                    # Parse error. 
                    orig_lineno = lineno

                    lineno += num_newlines
                    recno += 1
                    start_offset = end_offset

                    # See if I can finish off this record.
                    to_read = max(327680, 5*block_size) - len(block)
                    if to_read > 0:
                        block += infile.read(to_read)

                    i = block.find(b"\n$$$$\n", offset)
                    if i == -1:
                        end_offset += (new_offset-offset)
                        # Have a record for the Location
                        record = block[offset:]
                        error_handler.error("The text toolkit cannot parse the sdf record. It is "
                                            "not in the right format, and unable to continue",
                                            location)
                        return

                    # It looks like I can recover!
                    record = block[offset:i+6]
                    offset = i+6

                    # Update the location information and generate the error message
                    end_offset = start_offset + len(record)

                    error_handler.error("The text toolkit cannot parse the sdf record. It is "
                                        "not in the right format",
                                        location)

                    # Update to past the end of the bad record
                    lineno = orig_lineno + record.count(b"\n")

                    # Reset so a future get_record() doesn't return wrong information
                    record = None
                    
                    # Continue processing the current block
                    continue


                elif err == 5:
                    # Found a record, but without an id. (An empty id returns "".)
                    record = block[offset:new_offset]
                    #assert record.count(b"\n") == num_newlines, (record.count(b"\n"), num_newlines)

                    start_offset = end_offset
                    end_offset += (new_offset-offset)

                    recno += 1
                    yield (None, record)

                    # Keep track of where we are
                    lineno += num_newlines
                    offset = new_offset

                elif err == 6:
                    raise AssertionError("Argument error but I can't figure out what happened")

                else:
                    raise AssertionError("Unknown error code %r" % (err,))
    finally:
        # I seem to be confused about how 'lineno' works
        final_lineno = get_lineno()
        if final_lineno > 0:
            final_lineno -= 1
        location.save(lineno = final_lineno,
                      recno = get_recno(),
                      offsets = get_offsets(),
                      record = get_record())
        if close is not None:
            try:
                close()
            finally:
                close = None



def iter_sdf_ids_and_values(infile, close, id_tag_bytes, value_tag_bytes, error_handler, location,
                            encoding, encoding_errors, block_size = 327680):
    it = _iter_sdf_ids_and_values(infile, close, id_tag_bytes, value_tag_bytes,
                                  error_handler, location, block_size)
    next(it)
    def iter_sdf_ids_and_values():
        try:
            for id_bytes, value_bytes in it:
                if id_bytes is not None:
                    try:
                        id = id_bytes.decode(encoding, encoding_errors)
                    except UnicodeDecodeError as err:
                        # encoding_errors is 'strict', so report that as a true error
                        error_handler.error("Cannot decode the identifier %r: %s" % (id_bytes, err),
                                            location)
                        # I can recover from it by doing nothing
                        continue
                else:
                    id = None

                if value_bytes is not None:
                    try:
                        value = value_bytes.decode(encoding, encoding_errors)
                    except UnicodeDecodeError as err:
                        # encoding_errors is 'strict', so report that as a true error
                        error_handler.error("Cannot decode the value %r: %s" % (value_bytes, err),
                                            location)
                        # I can recover from it by doing nothing
                        continue
                else:
                    value = None

                yield id, value
        finally:
            it.close()
            
    return iter_sdf_ids_and_values()

def _iter_sdf_ids_and_values(infile, close, id_tag_bytes, value_tag_bytes,
                             error_handler, location, block_size):
    recno = 0
    lineno = 1
    start_offset = end_offset = offset = 0
    block = b""
    record = None

    def get_lineno():
        if recno == 0:
            return 0
        return lineno

    def get_recno():
        return recno

    def get_offsets():
        return start_offset, end_offset

    def get_record():
        return record

    location.register(get_lineno=get_lineno,
                      get_recno=get_recno,
                      get_offsets=get_offsets,
                      get_record=get_record)

    try:
        yield "Ready for processing!"
        
        while 1:
            # Read the next block.
            next_block = infile.read(block_size)

            if not next_block:
                # At the end of the file. Check for extra data.
                if not block:
                    return
                recno += 1
                record = block[offset:]
                error_handler.error("Incomplete sdf record or extra trailing data", location)
                return

            # Merge the current and next blocks
            block = block + next_block
            offset = 0

            while 1:
                # Figure out the contents and end of the current record.
                err, new_offset, id, value, num_newlines = _chemfp.find_record_two_fields(
                    block, offset, id_tag_bytes, value_tag_bytes)
                if err == 1:
                    # Found a record
                    record = block[offset:new_offset] # don't need this, but want error reporting
                    #assert record.count(b"\n") == num_newlines, (record.count(b"\n"), num_newlines)

                    # Update the location information then yield the id and record
                    start_offset = end_offset
                    end_offset += (new_offset-offset)

                    recno += 1
                    yield (id, value)


                    # Keep track of location information
                    lineno += num_newlines
                    offset = new_offset

                elif err == 2:
                    # No more content in this block. Reset, and fetch more data.
                    block = b""
                    break

                elif err == 3:
                    # Incomplete record. Need more data. Reset the block so it
                    # only contains the trailing data, and fetch more data.
                    block = block[offset:]
                    break

                elif err == 4:
                    # Parse error. Figure out location information.
                    orig_lineno = lineno

                    lineno += num_newlines
                    recno += 1
                    start_offset = end_offset

                    # See if I can finish off this record
                    to_read = max(327680, 5*block_size) - len(block)
                    if to_read > 0:
                        block += infile.read(to_read)

                    i = block.find(b"\n$$$$\n", offset)
                    if i == -1:
                        end_offset += (new_offset-offset)
                        record = block[offset:end_offset]
                        error_handler.error(
                            "The text toolkit cannot parse the sdf record. It is "
                            "not in the right format, and unable to continue",
                            location)
                        return

                    # It looks like I can recover!
                    record = block[offset:i+6]
                    offset = i+6

                    # Update the location information and generate the error message
                    end_offset = start_offset + len(record)

                    error_handler.error("The text toolkit cannot parse the sdf record. "
                                        "It is not in the right format", location)

                    # Update to past the end of the bad record
                    lineno = orig_lineno + record.count(b"\n")

                    # Reset so a future get_record() doesn't return wrong information
                    record = None

                    # Continue processing the current block
                    continue


                elif err == 5:
                    # Found a record, but without an id and/or without a value. (An empty id returns "".)
                    record = block[offset:new_offset]
                    #assert record.count(b"\n") == num_newlines, (record.count(b"\n"), num_newlines)

                    start_offset = end_offset
                    end_offset += (new_offset-offset)

                    recno += 1
                    yield (id, value)

                    # Keep track of where we are
                    lineno += num_newlines
                    offset = new_offset

                elif err == 6:
                    raise AssertionError("Argument error but I can't figure out what happened")

                else:
                    raise AssertionError("Unknown error code %r" % (err,))
    finally:
        # I seem to be confused about how 'lineno' works
        final_lineno = get_lineno()
        if final_lineno > 0:
            final_lineno -= 1
        location.save(lineno = final_lineno,
                      recno = get_recno(),
                      offsets = get_offsets(),
                      record = get_record())
        if close is not None:
            try:
                close()
            finally:
                close = None

def add_sdf_record_tag(sdf_record, tag, value):
    is_unicode = isinstance(sdf_record, unicode_type)
    # Some parameter validation so the error messages are consistent with 'get_tag'
    if isinstance(tag, unicode_type):
        tag_bytes = tag.encode("utf8")
        if not is_unicode:
            raise ValueError("cannot mix string and byte types for sdf_record and tag")
    elif isinstance(tag, bytes_type):
        tag_bytes = tag
        if is_unicode:
            raise ValueError("cannot mix string and byte types for sdf_record and tag")
    else:
        raise ValueError("tag %r must be a string or bytes" % (tag,)) # cannot use this to set the title
        
    tag_argument_error(tag_bytes)
    
    if isinstance(value, unicode_type):
        if not is_unicode:
            raise ValueError("cannot mix string and byte types for sdf_record and value")
        value_bytes = value.encode("utf8")
        
    elif isinstance(value, bytes_type):
        if is_unicode:
            raise ValueError("cannot mix string and byte types for sdf_record and value")
        value_bytes = value
    else:
        raise ValueError("value %r must be a string or bytes" % (value,))

    if b"\n\n" in value_bytes:
        raise ValueError("Value contains too many newlines")

    if is_unicode:
        if not sdf_record.endswith("\n$$$$\n"):
            raise ValueError("Cannot add tag because the record does not appear to be in SDF format")
        if value[-1:] == "\n":
            return sdf_record[:-5] + "> <%s>\n%s\n$$$$\n" % (tag, value)
        else:
            return sdf_record[:-5] + "> <%s>\n%s\n\n$$$$\n" % (tag, value)
    else:
        if not sdf_record.endswith(b"\n$$$$\n"):
            raise ValueError("Cannot add tag because the record does not appear to be in SDF format")
        if value[-1:] == b"\n":
            return sdf_record[:-5] + b"> <%s>\n%s\n$$$$\n" % (tag, value)
        else:
            return sdf_record[:-5] + b"> <%s>\n%s\n\n$$$$\n" % (tag, value)

    
_tag_pattern = re.compile(br">.*<([^>]*)>.*\n")
_next_tag_pattern = re.compile(br"(\n)\n(>|\$\$\$\$\n)")

def _get_sdf_tag_pairs_bytes(record):
    # Hack solution!
    start = record.find(b"M  END\n")
    if start == -1:
        return []
    
    start += 7
    if record[start:start+1] != b">":
        # At the '$$$$', or in the wrong format.
        return []

    terms = []
    while 1:
        m = _tag_pattern.match(record, start)
        if m is None:
            break
        tag = m.group(1)
        data_start = m.end()

        m = _next_tag_pattern.search(record, data_start)
        if m is None:
            break
        data_end = m.start(1)
        terms.append( (tag, record[data_start:data_end]) )

        start = m.start(2)

    return terms

def get_sdf_tag_pairs(record):
    if isinstance(record, bytes):
        return _get_sdf_tag_pairs_bytes(record)
    record = record.encode("utf8")
    return [(tag.decode("utf8"), value.decode("utf8")) for tag, value in _get_sdf_tag_pairs_bytes(record)]
    

## def get_sdf_tag_pairs(record, encoding="utf8", encoding_errors="strict"):
##     return [(tag.decode(encoding, encoding_errors),
##              value.decode(encoding, encoding_errors)) for (tag, value) in get_sdf_tag_pairs_as_bytes(record)]
    
def get_sdf_record_tag_as_bytes(record, tag):
    if isinstance(tag, unicode_type):
        raise ValueError("tag must be a byte string or None")
    elif isinstance(tag, bytes_type) or tag is None:
        pass
    else:
        raise ValueError("tag %r must be a string or bytes" % (tag,)) # otherwise there's a horrible error message
        
    err, new_offset, value, num_newlines = _chemfp.find_sdf_record_one_field(record, 0, tag)

    if err == 1:
        # Found the record
        return value

    elif err in (2, 3):
        raise ValueError("Incomplete SDF record")
    elif err == 4:
        raise ValueError("Record does not appear to be in SDF format")
    elif err == 5:
        # Found a value, though there's no id.
        return value
    elif err == 6:
        tag_argument_error(tag)
        raise AssertionError("Argument error but I can't figure out what happened")
    else:
        raise AssertionError("Unknown error code %r" % (err,))


def get_sdf_record_tag(record, tag, encoding="utf8", encoding_errors="strict"):
    if isinstance(tag, unicode_type):
        tag = tag.encode(encoding, encoding_errors)
    value = get_sdf_record_tag_as_bytes(record, tag)
    if value is None:
        return None
    return value.decode(encoding, encoding_errors)

####### SDF writers

def open_sdf_writer_gen(outfile, close, error_handler, location):
    it = _open_sdf_writer_gen(outfile, close, error_handler, location)
    next(it)
    return it

_valid_sdf_formats = set( (None, "sdf") )
def _open_sdf_writer_gen(outfile, close, error_handler, location):
                     
    lineno = recno = output_recno = 0
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    def get_lineno():
        # I track output newlines, but I'm not sure it's all that useful.
        return lineno

    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno,
                      get_lineno=get_lineno)
                      
    valid_formats = _valid_sdf_formats
    
    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecule":
                    id, rec = args
                    if id is None:
                        command = "write_molecules"
                        args = [rec]
                    else:
                        command = "write_ids_and_molecules"
                        args = [(id, rec)]

                if command == "write_molecules":
                    for rec in args:
                        recno += 1  # 'recno' is the number of *input* records
                        if rec.record_format not in valid_formats:
                            error_handler.error(
                                "Cannot write a record in %r format as SDF"
                                % (rec.record_format,), location)
                            continue

                        record = rec.record
                        
                        if record is None:
                            error_handler.error("Missing text record", location)
                            continue

                        num_newlines = record.count(b"\n")
                        if num_newlines < 6:
                            error_handler.error("Text record is too short to be a valid SDF record",
                                                location)
                            continue
                        if not record.endswith(b"$$$$\n"):
                            error_handler.error("Text record does not end with the expected '$$$$' line",
                                                location)
                            continue

                        id_bytes = rec.id_bytes
                        if id_bytes is None or record.startswith(id_bytes+b"\n"):
                            if not outfile.write_bytes(record, error_handler, location, rec.encoding, rec.encoding_errors):
                                continue
                        else:
                            if b"\n" in id_bytes:
                                error_handler.error("Id must not contain a newline", location)
                                continue
                            title, newline, remainder = record.partition(b"\n")
                            record = id_bytes + b"\n" + remainder
                            if not outfile.write_bytes(record, error_handler, location, rec.encoding, rec.encoding_errors):
                                continue

                        lineno += num_newlines
                        output_recno += 1

                elif command == "write_ids_and_molecules":
                    for id, rec in args:
                        recno += 1  # 'recno' is the number of *input* records
                        if rec.record_format not in valid_formats:
                            error_handler.error(
                                "Cannot write a record in %r format as SDF" % (rec.record_format,), location)
                            continue

                        record = rec.record
                        if record is None:
                            error_handler.error("Missing text record", location)
                            continue

                        num_newlines = record.count(b"\n")
                        if num_newlines < 6:
                            error_handler.error("Text record is too short to be a valid SDF record", location)
                            continue
                        if not record.endswith(b"$$$$\n"):
                            error_handler.error("Text record does not end with the expected '$$$$' line",
                                                location)
                            continue
                        
                        if id is None:
                            id = rec.id
                        
                        if id is None:
                            if not outfile.write_bytes(record, error_handler, location, rec.encoding, rec.encoding_errors):
                                continue
                        else:
                            if "\n" in id:
                                error_handler.error("Id must not contain a newline", location)
                                continue
                            i = record.index(b"\n")
                            if not outfile.write_id_then_remainder(id, record[i:], error_handler, location,
                                                            rec.encoding, rec.encoding_errors):
                                continue

                        lineno += num_newlines
                        output_recno += 1

                else:
                    raise AssertionError("Unknown command: %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
                
    finally:
        location.save(recno = get_recno(),
                      output_recno = get_output_recno(),
                      lineno = get_lineno())
        if close is not None:
            try:
                close()
            finally:
                close = None
            
                    
            
    

#########

def _to_sdf_record(reader_iter, location, encoding, encoding_errors):
    it = _to_sdf_record_iter(reader_iter, location, encoding, encoding_errors)
    next(it)
    return it
    
def _to_sdf_record_iter(reader_iter, location, encoding, encoding_errors):
    mol = None
    def get_mol():
        return mol
    location.register(get_mol=get_mol)
    try:
        yield "Ready for processing!"
        for id, content in reader_iter:
            # Figure out the actual title
            title = content[:content.index(b"\n")]
            if title[-1:] == b"\r":
                title = title[:-1]
            mol = SDFRecord(title, content, encoding, encoding_errors)
            yield id, mol
            mol = None
    finally:
        mol = None
        location.save(mol=None)
        reader_iter.close()



class TextFormatConfig(object):
    is_available = True
    alternate_extensions = ()
    class input(object):
        arg_info = (
            ("delimiter", base_toolkit.parse_delimiter),
            ("has_header", base_toolkit.parse_boolean),
            )
        defaults = {
            "delimiter": None,
            "has_header": False,
            }
    
    class output(object):
        arg_info = (
            ("delimiter", base_toolkit.parse_delimiter),
            )
        defaults = {
            "delimiter": None,
            }

def _make_smi_records(reader_iter, location, record_klass, encoding, encoding_errors):
    mol = None
    def get_mol():
        return mol
    location.register(get_mol=get_mol)
    try:
        for line, smi, id in reader_iter:
            mol = record_klass(id, line, smi, encoding, encoding_errors)
            yield id, mol
    finally:
        location.save(mol=None)

class BaseSmiFormat(object):
    supports_io = True
    _valid_formats = set((None, "smi", "can", "usm", "smistring", "canstring", "usmstring"))
    
    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        has_header = bool(reader_args["has_header"])

        infile, close = io.open_compressed_input(source, compression, BLAME_NAME)
    
        reader_iter = iter_line_and_first_two_delimited_fields_as_strings(
            infile, close, delimiter, has_header, error_handler,
            location, self.register_name, encoding, encoding_errors)
        reader_iter = _make_smi_records(reader_iter, location, self.record_klass, encoding, encoding_errors)
        return base_toolkit.IdAndMoleculeReader(
            base_toolkit.FormatMetadata(location.filename, self.record_format,
                                        {"delimiter": delimiter.name, "has_header": has_header}),
            reader_iter,
            location)

    def read_ids_and_molecules_from_string(self, content, compression, id_tag, reader_args,
                                           error_handler, location):
        is_unicode, infile, close = io._open_string_as_bytes(content, compression, BLAME_NAME)
        encoding = "utf8"
        encoding_errors = "strict"
                    
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        has_header = bool(reader_args["has_header"])

        reader_iter = iter_line_and_first_two_delimited_fields_as_strings(
            infile, close, delimiter, has_header, error_handler,
            location, self.register_name,
            encoding, encoding_errors)
        reader_iter = _make_smi_records(reader_iter, location, self.record_klass, encoding, encoding_errors)
        
        return base_toolkit.IdAndMoleculeReader(
            base_toolkit.FormatMetadata(location.filename, self.record_format,
                                        {"delimiter": delimiter.name, "has_header": has_header}),
            reader_iter,
            location)

    def make_id_and_molecule_parser(self, id_tag_bytes, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        split_two_terms_string = delimiter.split_two_terms_string
        split_two_terms_bytes = delimiter.split_two_terms_bytes
        record_klass = self.record_klass

        def text_parser(content):
            if isinstance(content, unicode_type):
                line, newline, rest = content.partition("\n")  # What about \r?
                terms = split_two_terms_string(line)
                if not terms:
                    error_handler.error("The empty string is not a valid SMILES input")
                    return None, None
                smiles, id = terms
                if not smiles:
                    error_handler.error("The empty string is not a valid SMILES input")
                    return None, None
                return id, record_klass(id, line + newline, smiles, "utf8", "strict")
            else:
                line, newline, rest = content.partition(b"\n")  # What about \r?
                terms = split_two_terms_bytes(line)
                if not terms:
                    error_handler.error("The empty string is not a valid SMILES input")
                    return None, None
                smiles_bytes, id_bytes = terms
                if not smiles_bytes:
                    error_handler.error("The empty string is not a valid SMILES input")
                    return None, None
                try:
                    smiles = smiles_bytes.decode("utf8")
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the SMILES %r: %s" % (smiles_bytes, err))
                    return None, None
                
                if id_bytes is None:
                    id = None
                else:
                    try:
                        id = id_bytes.decode("utf8")
                    except UnicodeDecodeError as err:
                        error_handler.error("Cannot decode the identifier %r: %s" % (id_bytes, err))
                        return None, None
                
                return id, record_klass(id, line + newline, smiles, "utf8", "strict")

        text_parser.__name__ = "text_" + self.register_name + "_parse_id_and_molecule"
        return text_parser
    

    def parse_id_and_molecule(self, content, id_tag_bytes, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        if isinstance(content, unicode_type):
            record, newline, rest = content.partition("\n")
            terms = delimiter.split_two_terms_string(record)
            if not terms:
                error_handler.error("The empty string is not a valid SMILES input")
                return None, None
            smiles, id = terms
            if not smiles:
                error_handler.error("The empty string is not a valid SMILES input")
                return None, None
            record_bytes = (record+newline).encode("utf8")
        else:
            record_bytes, newline, rest = content.partition(b"\n")
            terms = delimiter.split_two_terms_bytes(record_bytes)
            if not terms:
                error_handler.error("The empty string is not a valid SMILES input")
                return None, None
            smiles_bytes, id_bytes = terms
            record_bytes += newline
            
            if not smiles_bytes:
                error_handler.error("The empty string is not a valid SMILES input")
                return None, None
            try:
                smiles = smiles_bytes.decode("utf8")
            except UnicodeDecodeError as err:
                error_handler.error("Cannot decode the SMILES %r: %s" % (smiles_bytes, err))
                return None, None
            
            if id_bytes is None:
                id = None
            else:
                try:
                    id = id_bytes.decode("utf8")
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the identifier %r: %s" % (id_bytes, err))
                    return None, None
                
        
        return id, self.record_klass(id, record_bytes, smiles, "utf8", "strict")

    def open_molecule_writer(self, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, writer_args)

        return base_toolkit.MoleculeWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format,
                                        {"delimiter": delimiter.name}),
            open_two_delimited_fields_writer_gen(
                outfile, close, delimiter, error_handler, location, "SMILES", BaseSmiFormat._valid_formats,
                encoding, encoding_errors),
            location)
            
    def open_molecule_writer_to_string(self, writer_args, errors, location):
        delimiter = base_toolkit.get_delimiter(self, writer_args)

        outfile, close, getvalue = io._open_output_string()
        
        return base_toolkit.MoleculeStringWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format,
                                        {"delimiter": delimiter.name}),
            open_two_delimited_fields_writer_gen(
                outfile, close, delimiter, errors, location, "SMILES", BaseSmiFormat._valid_formats,
                None, None),
            getvalue,
            location)

    def open_molecule_writer_to_bytes(self, compression, writer_args, errors, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args)

        outfile, close, getvalue = io._open_compressed_output_bytes(compression, BLAME_NAME, level)
        
        return base_toolkit.MoleculeStringWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format,
                                        {"delimiter": delimiter.name}),
            open_two_delimited_fields_writer_gen(
                outfile, close, delimiter, errors, location, "SMILES", BaseSmiFormat._valid_formats,
                "utf8", "strict"),
            getvalue,
            location)
                    

    def create_string(self, mol, id, writer_args, errors):
        #delimiter = base_toolkit.get_delimiter(self, writer_args)
        if mol.record_format not in self._valid_formats:
            raise ValueError("Cannot convert a %r record to %r format" %
                             (mol.record_format, self.record_format))
        return _create_delimited_smi_string(self, mol, id, writer_args, errors)

    def create_bytes(self, mol, id, writer_args, error_handler):
        return self.create_string(mol, id, writer_args, error_handler).encode("utf8")


class SmiFormat(TextFormatConfig, BaseSmiFormat):
    register_name = record_format = "smi"
    alternate_extensions = ("ism", "isosmi")
    record_klass = SmiRecord

class CanFormat(TextFormatConfig, BaseSmiFormat):
    register_name = record_format = "can"
    record_klass = CanRecord
    
class UsmFormat(TextFormatConfig, BaseSmiFormat):
    register_name = record_format = "usm"
    record_klass = UsmRecord

class SDFFormat(TextFormatConfig):
    register_name = record_format = "sdf"
    alternate_extensions = ("sd", "mdl")
    supports_io = True
    record_klass = SDFRecord
    
    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        id_tag_bytes = get_id_tag_bytes(id_tag, encoding, encoding_errors)
        infile, close = io.open_compressed_input(source, compression, BLAME_NAME)
        reader_iter = iter_sdf_ids_and_records(infile, close, id_tag_bytes, error_handler, location,
                                               encoding, encoding_errors)
        reader_iter = _to_sdf_record(reader_iter, location, encoding, encoding_errors)
        return base_toolkit.IdAndMoleculeReader(
            base_toolkit.FormatMetadata(location.filename, "sdf", reader_args),
            reader_iter,
            location)


    def read_ids_and_molecules_from_string(self, content, compression, id_tag, reader_args,
                                           error_handler, location):
        is_unicode, infile, close = io._open_string_as_bytes(content, compression, BLAME_NAME)
        if is_unicode:
            if isinstance(id_tag, bytes_type):
                raise ValueError("id_tag must be a string, not bytes, when the content is a string")
        encoding = "utf8"
        encoding_errors = "strict"

        id_tag_bytes = get_id_tag_bytes(id_tag, encoding, encoding_errors)
        reader_iter = iter_sdf_ids_and_records(infile, close, id_tag_bytes, error_handler, location,
                                               encoding, encoding_errors)
        reader_iter = _to_sdf_record(reader_iter, location, encoding, encoding_errors)
        return base_toolkit.IdAndMoleculeReader(
            base_toolkit.FormatMetadata(location.filename, "sdf", reader_args),
            reader_iter,
            location)


    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        if not content:
            error_handler.error("Cannot parse a record from an empty string")
            return None, None
        if isinstance(content, unicode_type):
            content = content.encode("utf8")
        id_tag_bytes = get_id_tag_bytes(id_tag, "utf8", "strict")

        err, new_offset, id_bytes, num_newlines = _chemfp.find_sdf_record_one_field(content, 0, id_tag_bytes)
        if err == 1:
            # Found a record. Get the title, and return the molecule record
            title = content[:content.index(b"\n")]
            if title[-1:] == b"\r":
                title = title[:-1]
            try:
                id = id_bytes.decode("utf8")
            except UnicodeDecodeError as err:
                # encoding_errors is 'strict', so report that as a true error
                error_handler.error("Cannot decode the identifier %r: %s" % (id_bytes, err))
                return None, None

            return id, SDFRecord(title, content[:new_offset], "utf8", "strict")
        elif err == 2:
            error_handler.error("The text toolkit cannot parse the sdf record. Empty record.")
        elif err == 3:
            error_handler.error("The text toolkit cannot parse the sdf record. It appears to be incomplete.")
        elif err == 4:
            error_handler.error("The text toolkit cannot parse the sdf record. "
                                "Incorrect format at line %d" % (num_newlines,))
        elif err == 5:
            # record without an id. Get the title, and return the molecule record
            # TODO: push this into a C function
            title = content[:content.index(b"\n")]
            if title[-1:] == b"\r":
                title = title[:-1]

            return None, SDFRecord(title, content[:new_offset], "utf8", "strict")
        else:
            raise AssertionError("Unexpected error code %d" % (err,))

        # Can get here when there was an error in the parser and the error_handler
        # said to ignore it
        return None, None

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        id_tag_bytes = get_id_tag_bytes(id_tag, "utf8", "strict")
            
        def text_sdf_parse_id_and_molecule(content):
            if not content:
                error_handler.error("Cannot parse a record from an empty string")
                return None, None

            if isinstance(content, unicode_type):
                content = content.encode("utf8", "strict")
            
            err, new_offset, id_bytes, num_newlines = _chemfp.find_sdf_record_one_field(content, 0, id_tag_bytes)
            if err == 1:
                # Found a record. Get the title, and return the molecule record
                offset = content.index(b"\n")
                title = content[:offset]
                if title[-1:] == b"\r":
                    title = title[:-1]
                try:
                    id = id_bytes.decode("utf8")
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the identifier %r: %s" % (id_bytes, err))
                    return None, None
                else:
                    return id, SDFRecord(title, content[:new_offset], "utf8", "strict")
            elif err == 5:
                # record without an id. Get the title, and return the molecule record
                offset = content.index(b"\n")
                title = content[:offset]
                if title[-1:] == b"\r":
                    title = title[:-1]
                
                return None, SDFRecord(title, content[:new_offset], "utf8", "strict")

            elif err == 2:
                error_handler.error("The text toolkit cannot parse the sdf record. Empty record.")
            elif err == 3:
                error_handler.error("The text toolkit cannot parse the sdf record. It appears to be incomplete.")
            elif err == 4:
                error_handler.error("The text toolkit cannot parse the sdf record. "
                                    "Incorrect format at line %d" % (num_newlines,))
            else:
                raise AssertionError("Unexpected error code %d" % (err,))

            # fall-through in case error_handler is 'ignore'
            return None, None

        return text_sdf_parse_id_and_molecule

        
    def open_molecule_writer(self, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        return base_toolkit.MoleculeWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format, writer_args),
            open_sdf_writer_gen(outfile, close, error_handler, location),
            location)

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        #outfile, close, getvalue = _open_to_string()
        outfile, close, getvalue = io._open_output_string()
        return base_toolkit.MoleculeStringWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format, writer_args),
            open_sdf_writer_gen(outfile, close, error_handler, location),
            getvalue,
            location)
    
    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        outfile, close, getvalue = io._open_compressed_output_bytes(compression, BLAME_NAME, level)
        return base_toolkit.MoleculeStringWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format, writer_args),
            open_sdf_writer_gen(outfile, close, error_handler, location),
            getvalue,
            location)

    def create_string(self, mol, id, writer_args, error_handler):
        record = mol.record
        try:
            record = record.decode(mol.encoding, mol.encoding_errors)
        except UnicodeDecodeError as err:
            error_handler.error("Cannot decode the record with id %r: %s" % (mol.id, err))
            return None
        
        if id is None:
            id = mol.id
            if id is None:
                return record

        if u"\n" in id:
            raise ValueError("Id must not contain a newline")
        
        title, newline, remainder = record.partition("\n")
        return id + newline + remainder

    def create_bytes(self, mol, id, writer_args, error_handler):
        record = mol.record
        if mol.encoding == "utf8":
            record_bytes = record
            if id is None:
                id_bytes = mol.id_bytes
            else:
                try:
                    id_bytes = id.encode("utf8")
                except UnicodeEncodeError as err:
                    error_handler.error("Cannot encode the identifier %r: %s" % (id, err))
                    return None
            if b"\n" in id_bytes:
                raise ValueError("Id must not contain a newline")
                
            if id_bytes is not None and not record_bytes.startswith(id_bytes + b"\n"):
                title, newline, remainder = record_bytes.partition(b"\n")
                record_bytes = id_bytes + newline + remainder
        
        else:
            # Transcode
            try:
                record_str = record.decode(mol.encoding, mol.encoding_errors)
            except UnicodeDecodeError as err:
                error_handler.error("Cannot decode the record with id %r: %s" % (mol.id, err))
                return None

            if id is None:
                id = mol.id
            if id is not None and not record_str.startswith(id + "\n"):
                title, newline, remainder = record_str.partition("\n")
                record_str = id + newline + remainder
            
            record_bytes = record_str.encode("utf8")
        
        return record_bytes
        

class BaseSmilesStringFormat(base_toolkit.NotIoFormatConfigMixin, TextFormatConfig):
    supports_io = False
    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        record_klass = self.record_klass
        def text_smiles_parser(content):
            # Follow the Daylight spec, which says that the SMILES ends with any whitespace
            terms = content.split(None, 1)
            if not terms:
                # In chemfp, the empty SMILES string returns an empty molecule.
                return None, record_klass(None, b"", "")

            smiles = terms[0]
            if not isinstance(smiles, unicode_type):
                smiles_bytes = smiles
                try:
                    smiles = smiles_bytes.decode("utf8")
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the SMILES %r: %s" % (smiles_bytes, err))
                    return None, None
            return None, record_klass(None, smiles.encode("utf8"), smiles)

        text_smiles_parser.__name__ = "text_" + self.register_name + "_parse_id_and_molecule"
        return text_smiles_parser
    
    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        # Follow the Daylight spec, which says that the SMILES ends with any whitespace
        terms = content.split(None, 1)
        if not terms:
            # In chemfp, the empty SMILES string returns an empty molecule.
            return None, self.record_klass(None, b"", "")

        smiles = terms[0]
        if not isinstance(smiles, unicode_type):
            smiles_bytes = smiles
            try:
                smiles = smiles_bytes.decode("utf8")
            except UnicodeDecodeError as err:
                error_handler.error("Cannot decode the SMILES %r: %s" % (smiles_bytes, err))
                return None, None
        else:
            smiles_bytes = smiles.encode("utf8")
        
        return None, self.record_klass(None, smiles_bytes, smiles)

    def create_string(self, mol, id, writer_args, error_handler):
        return mol.smiles

    def create_bytes(self, mol, id, writer_args, error_handler):
        return mol.smiles.encode("utf8")

class SmiStringFormat(BaseSmilesStringFormat):
    register_name = record_format = "smistring"
    record_klass = SmiStringRecord

class CanStringFormat(BaseSmilesStringFormat):
    register_name = record_format = "canstring"
    record_klass = CanStringRecord

class UsmStringFormat(BaseSmilesStringFormat):
    register_name = record_format = "usmstring"
    record_klass = UsmStringRecord


# Low-level SDF record readers

def _get_sdf_compression(source, compression):
    if compression == "gz":
        return "gz"
    elif compression in ("", "none"):
        return ""
    elif compression in (None, "auto"):
        ext, file_compression = io.normalize_input_format(source, None, ("sdf", ""))
        if file_compression in ("", "gz", "zst", "bz2"):
            return file_compression
        raise ValueError("The text toolkit does not support %r compression" % (file_compression,))
    elif compression == "zst":
        return "zst"
    else:
        raise ValueError("The text toolkit does not support %r compression" % (compression,))
    

def read_sdf_records(source=None, reader_args=None, compression="auto",
                     errors="strict", location=None, block_size=327680):
    compression = _get_sdf_compression(source, compression)

    error_handler = io.get_parse_error_handler(errors)

    # Track location
    if location is None:
        location = io.Location.from_source(source)
    location.save(record_format="sdf")

    infile, close = io.open_compressed_input(source, compression, "The text toolkit")
    reader_iter = iter_sdf_records(infile, close, error_handler, location, block_size)
    return base_toolkit.RecordReader(
        base_toolkit.FormatMetadata(location.filename, "sdf", {}),
        reader_iter,
        location)
    

def read_sdf_ids_and_records(source=None, id_tag=None, reader_args=None, compression="auto",
                            errors="strict", location=None,
                            encoding="utf8", encoding_errors="strict", block_size=327680):
    compression = _get_sdf_compression(source, compression)

    id_tag_bytes = get_id_tag_bytes(id_tag, encoding, encoding_errors)
    _check_block_size(block_size)

    error_handler = io.get_parse_error_handler(errors)

    # Track location
    if location is None:
        location = io.Location.from_source(source)
    location.save(record_format="sdf")

    infile, close = io.open_compressed_input(source, compression, "The text toolkit")
    reader_iter = iter_sdf_ids_and_records(infile, close, id_tag_bytes, error_handler, location,
                                           encoding, encoding_errors, block_size)
    return base_toolkit.IdAndRecordReader(
        base_toolkit.FormatMetadata(location.filename, "sdf", {}),
        reader_iter,
        location)
    

def read_sdf_ids_and_values(source=None, id_tag=None, value_tag=None, reader_args=None, compression="auto",
                            errors="strict", location=None,
                            encoding="utf8", encoding_errors="strict", block_size=327680):
    compression = _get_sdf_compression(source, compression)

    id_tag_bytes = get_id_tag_bytes(id_tag, encoding, encoding_errors)
    value_tag_bytes = get_value_tag_bytes(value_tag, encoding, encoding_errors)
    _check_block_size(block_size)

    error_handler = io.get_parse_error_handler(errors)

    # Track location
    if location is None:
        location = io.Location.from_source(source)
    location.save(record_format="sdf")

    infile, close = io.open_compressed_input(source, compression, "The text toolkit")
    reader_iter = iter_sdf_ids_and_values(infile, close, id_tag_bytes, value_tag_bytes,
                                          error_handler, location, encoding, encoding_errors, block_size)
    return base_toolkit.IdAndRecordReader(
        base_toolkit.FormatMetadata(location.filename, "sdf", {}),
        reader_iter,
        location)

            

def read_sdf_records_from_string(content, reader_args=None, compression=None,
                                 errors="strict", location=None, block_size=327680):
    _check_block_size(block_size)

    error_handler = io.get_parse_error_handler(errors)

    # Track location
    if location is None:
        location = io.Location.from_source("<string>")
    location.save(record_format="sdf")


    is_unicode, infile, close = io._open_string_as_bytes(content, compression, BLAME_NAME)

    # Ignore id encoding failures because I'm going to ignore it anyway.
    reader_iter = iter_sdf_ids_and_records(infile, close, None, error_handler, location,
                                           "utf8", "ignore", block_size)
    if is_unicode:
        def read_sdf_records_from_string(reader_iter):
            for id_bytes, record_bytes in reader_iter:
                yield record_bytes.decode("utf8")
    else:
        def read_sdf_records_from_string(reader_iter):
            for id_bytes, record_bytes in reader_iter:
                yield record_bytes

    return base_toolkit.RecordReader(
        base_toolkit.FormatMetadata(location.filename, "sdf", {}),
        read_sdf_records_from_string(reader_iter),
        location)



def read_sdf_ids_and_records_from_string(content=None, id_tag=None, reader_args=None, compression=None,
                                        errors="strict", location=None,
                                        encoding="utf8", encoding_errors="strict", block_size=327680):
    _check_block_size(block_size)
    is_unicode, infile, close = io._open_string_as_bytes(content, compression, BLAME_NAME)

    if is_unicode:
        if isinstance(id_tag, bytes_type):
            raise ValueError("id_tag must be a string, not bytes, when the content is a string")
        encoding = "utf8"
        encoding_errors = "strict"

    id_tag_bytes = get_id_tag_bytes(id_tag, encoding, encoding_errors)

    error_handler = io.get_parse_error_handler(errors)

    # Track location
    if location is None:
        location = io.Location.from_source("<string>")
    location.save(record_format="sdf")


    reader_iter = iter_sdf_ids_and_records(infile, None, id_tag_bytes, error_handler, location,
                                           encoding, encoding_errors, block_size)
    if is_unicode:
        reader_iter = ((id, record_bytes.decode("utf8")) for (id, record_bytes) in reader_iter)
                            
    return base_toolkit.IdAndRecordReader(
        base_toolkit.FormatMetadata(location.filename, "sdf", {}),
        reader_iter,
        location)



def read_sdf_ids_and_values_from_string(content=None, id_tag=None, value_tag=None, compression=None,
                                        reader_args=None, errors="strict", location=None,
                                        encoding="utf8", encoding_errors="strict", block_size=327680):
    _check_block_size(block_size)
    is_unicode, infile, close = io._open_string_as_bytes(content, compression, BLAME_NAME)
    if is_unicode:
        if isinstance(id_tag, bytes_type):
            raise ValueError("id_tag must be a string, not bytes, when the content is a string")
        if isinstance(id_tag, bytes_type):
            raise ValueError("value_tag must be a string, not bytes, when the content is a string")
        encoding = "utf8"
        encoding_errors = "strict"
        
    id_tag_bytes = get_id_tag_bytes(id_tag, encoding, encoding_errors)
    value_tag_bytes = get_value_tag_bytes(value_tag, encoding, encoding_errors)
    
    error_handler = io.get_parse_error_handler(errors)

    # Track location
    if location is None:
        location = io.Location.from_source("<string>")
    location.save(record_format="value")

    reader_iter = iter_sdf_ids_and_values(infile, close, id_tag_bytes, value_tag_bytes,
                                          error_handler, location, encoding, encoding_errors, block_size)

    return base_toolkit.IdAndRecordReader(
        base_toolkit.FormatMetadata(location.filename, "sdf", {}),
        reader_iter,
        location)


# The format registry and toolkit
    
format_registry = base_toolkit.FormatRegistry("The text toolkit", "text")

format_registry.register_formats(
    SmiFormat(),
    CanFormat(),
    UsmFormat(),

    SmiStringFormat(),
    CanStringFormat(),
    UsmStringFormat(),
    
    SDFFormat(),
    )


_known_encodings = {}


def _check_valid_encoding(encoding, encoding_errors):
    key = encoding, encoding_errors
    try:
        return _known_encodings[key]
    except KeyError:
        pass
    
    from encodings import codecs
    try:
        codecs.lookup(encoding)
    except LookupError:
        v = _known_encodings[key] = "does not support encoding %r" % (encoding,)
        return v
    
    try:
        codecs.lookup_error(encoding_errors)
    except LookupError:
        v = _known_encodings[key] = "does not support encoding_errors %r" % (encoding,)
        return v

    # Prevent unbounded growth
    if len(_known_encodings) > 100:
        _known_encodings.clear()

    _known_encodings[key] = None
    return None


class TextToolkit(base_toolkit.BaseToolkit):
    def _check_read_encoding(self, format, encoding, encoding_errors):
        errmsg = _check_valid_encoding(encoding, encoding_errors)
        if errmsg is None:
            return

        raise ValueError("Format %s %s" % (format.prefix, errmsg))

    _check_write_encoding = _check_read_encoding
    
    def _open_molecule_writer_impl(self, format_config, destination, compression,
                                   writer_args, error_handler, location,
                                   encoding, encoding_errors, level):
        # Open the file and pass to the format
        outfile, close = io.open_compressed_output_wrapped(destination, compression, BLAME_NAME,
                                                           encoding, encoding_errors, level)

        return format_config.open_molecule_writer(outfile, close, writer_args, error_handler, location,
                                                  encoding, encoding_errors)


toolkit = TextToolkit(format_registry)
