"Internal support for the toolkit APIs"

# Copyright (c) 2010-2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import absolute_import

from . import io, _compat, EncodingError

# Base format and toolkit API


####### Deal with source/destination readers/writers

# The following are parts of the public API

class FormatMetadata(object):
    """Information about the reader or writer

    The public attributes are:

    .. py:attribute:: filename

       the source or destination filename, the string "<string>" for
       string-based I/O, or None if not known
        
    .. py:attribute:: record_format

       the normalized record format name. All SMILES formats are "smi",
       and this does not contain compression information

    .. py:attribute:: args

       the final reader_args or writer_args, after all processing,
       and as used by the reader and writer
    """
    def __init__(self, filename, record_format, args):
        self.filename = filename
        #assert record_format not in ("ism", "can")
        self.record_format = record_format  # Normalized, so "smi" for ism, can and smi formats
        self.args = args   # the reader or writer args

    def __repr__(self):
      """Return a string like 'FormatMeta(filename="cmpds.sdf.gz", record_format="sdf", args={})'"""
      return "FormatMetadata(filename=%r, record_format=%r, args=%r)" % (
        self.filename, self.record_format, self.args)
        

# The various readers all work the same way, except that iter(reader)
# returns different things. Formalize that with different subclasses.

class BaseMoleculeReader(object):
    """Base class for the toolkit readers

    The public attributes are:

    .. py:attribute:: metadata

       a :class:`chemfp.base_toolkit.FormatMetadata` instance

    .. py:attribute:: location

       a :class:`chemfp.io.Location` instance

    .. py:attribute:: closed

       False if the reader is open, otherwise True

    Readers are iterators, so iter(reader) returns itself.
    next(reader) returns either a single object or a pair of
    objects depending on reader.

    Readers are also a context manager, and call self.close()
    during exit.
    """
    def __init__(self, metadata, structure_reader, location):
        """(This constructor is not part of the public API)"""
        self.metadata = metadata
        self._structure_reader = structure_reader
        assert structure_reader is not None
        assert location is not None
        self.location = location
        self.closed = False

    def __enter__(self):
        """The reader supports the context manager API"""
        return self

    def __exit__(self, *args):
        """The reader context manager calls close() on exit"""
        self.close()

    def __iter__(self):
        return self._structure_reader

    def __next__(self):
        return next(self._structure_reader)

    def next(self):
        # This is for Python 2
        return self._structure_reader.next()

    def close(self):
        """Close the reader

        If the reader wasn't previously closed then close it. This will
        set the location properties to their final values, close any
        files that the reader may have opened, and set ``self.closed`` to False.
        """
        if self.closed:
            return
        self.closed = True
        self._structure_reader.close()
        
class IdAndMoleculeReader(BaseMoleculeReader):
    """Read structures from a file and iterate over the (id, toolkit molecule) pairs

    The public attributes are:

    .. py:attribute:: metadata

       a :class:`chemfp.base_toolkit.FormatMetadata` instance

    .. py:attribute:: location

       a :class:`chemfp.io.Location` instance

    .. py:attribute:: closed

       False if the reader is open, otherwise True

    Note: the toolkit implementation is free to reuse a molecule
    instead of returning a new one each time.
    """
    pass

class IdAndRecordReader(BaseMoleculeReader):
    """Read records from file and iterate over the (id, record string) pairs

    The public attributes are:

    .. py:attribute:: metadata

       a :class:`chemfp.base_toolkit.FormatMetadata` instance

    .. py:attribute:: location

       a :class:`chemfp.io.Location` instance

    .. py:attribute:: closed

       False if the reader is open, otherwise True
    """
    pass
    
class MoleculeReader(BaseMoleculeReader):
    """Read structures from a file and iterate over the toolkit molecules

    The public attributes are:

    .. py:attribute:: metadata

       a :class:`chemfp.base_toolkit.FormatMetadata` instance

    .. py:attribute:: location

       a :class:`chemfp.io.Location` instance

    .. py:attribute:: closed

       False if the reader is open, otherwise True

    Note: the toolkit implementation is free to reuse a molecule instead
    of returning a new one each time.
    """
    pass

class RecordReader(BaseMoleculeReader):
    """Read and iterate over records as strings

    The public attributes are:

    .. py:attribute:: metadata

       a :class:`chemfp.base_toolkit.FormatMetadata` instance

    .. py:attribute:: location

       a :class:`chemfp.io.Location` instance

    .. py:attribute:: closed

       False if the reader is open, otherwise True
    """
    pass

###

# The file writer and string writer work in much the same way.
# The base class implements the common core.

class BaseMoleculeWriter(object):
    """The base molecule writer API, implemented by :class:`MoleculeWriter` and :class:`MoleculeStringWriter`

    The public attributes are:

    .. py:attribute:: metadata

       a :class:`chemfp.base_toolkit.FormatMetadata` instance

    .. py:attribute:: location

       a :class:`chemfp.io.Location` instance

    .. py:attribute:: closed

       False if the reader is open, otherwise True

    The writer is a context manager, which calls self.close() when
    the manager exits.
    """
    def __init__(self, metadata, structure_writer, location):
        """(This constructor is not part of the public API)"""
        self.metadata = metadata
        self._structure_writer = structure_writer
        self.location = location
        self.closed = False
    
    def close(self):
        """Close the writer

        If the reader wasn't previously closed then close it. This will
        set the location properties to their final values, close any
        files that the writer may have opened, and set ``self.closed`` to False.
        """
        if self.closed:
            return
        self.closed = True
        self._structure_writer.close()

    def write_molecule(self, mol):
        """Write a toolkit molecule

        :param mol: the molecule to write
        :type mol: a toolkit molecule
        """
        if self.closed:
            raise ValueError("Cannot add a molecule after calling close")
        err = self._structure_writer.send( ("write_molecule", (None, mol)) )
        if err is not None:
            _compat.raise_tb(err[0], err[1])

    def write_molecules(self, mols):
        """Write a sequence of molecules

        :param mols: the molecules to write
        :type mols: a toolkit molecule iterator
        """
        if self.closed:
            raise ValueError("Cannot add a molecule after calling close")
        err = self._structure_writer.send( ("write_molecules", mols) )
        if err is not None:
            _compat.raise_tb(err[0], err[1])

    def write_id_and_molecule(self, id, mol):
        """Write an identifier and toolkit molecule

        If id is None then the output uses the molecule's own id/title.
        Specifying the id may modify the molecule's id/title, depending
        on the format and toolkit.

        :param id: the identifier to use for the molecule
        :type id: string, or None
        :param mol: the molecule to write
        :type mol: a toolkit molecule
        """
        if self.closed:
            raise ValueError("Cannot add a molecule after calling close")
        err = self._structure_writer.send( ("write_molecule", (id, mol)) )
        if err is not None:
            _compat.raise_tb(err[0], err[1])

    def write_ids_and_molecules(self, ids_and_mols):
        """Write a sequence of (id, molecule) pairs

        This function works well with :func:`chemfp.toolkit.read_ids_and_molecules()`,
        for example, to convert an SD file to SMILES file, and use an
        alternate *id_tag* to specify an alternative identifier.

        :param mols: the molecules to write
        :type mols: a (id string, toolkit molecule) iterator
        """
        if self.closed:
            raise ValueError("Cannot add a molecule after calling close")
        err = self._structure_writer.send( ("write_ids_and_molecules", ids_and_mols) )
        if err is not None:
            _compat.raise_tb(err[0], err[1])

    def __enter__(self):
        """The writer supports the context manager API"""
        return self

    def __exit__(self, *args):
        """The writer context manager calls close() on exit"""
        self.close()


class MoleculeWriter(BaseMoleculeWriter):
    """A BaseMoleculeWriter which writes molecules to a file.

    The public attributetes are:

    .. py:attribute:: metadata

       a :class:`chemfp.base_toolkit.FormatMetadata` instance

    .. py:attribute:: location

       a :class:`chemfp.io.Location` instance

    .. py:attribute:: closed

       False if the reader is open, otherwise True

    The writer is a context manager, which calls self.close() when
    the manager exits.
    """
    pass


class MoleculeStringWriter(BaseMoleculeWriter):
    """A BaseMoleculeWriter which writes molecules to a string.

    The public properties are:

    .. py:attribute:: metadata

       a :class:`chemfp.base_toolkit.FormatMetadata` instance

    .. py:attribute:: location

       a :class:`chemfp.io.Location` instance

    .. py:attribute:: closed

       False if the reader is open, otherwise True

    The writer is a context manager, which calls self.close() when
    the manager exits.
    """
    def __init__(self, details, structure_writer, getvalue, location):
        assert details is not None
        BaseMoleculeWriter.__init__(self, details, structure_writer, location)
        self._getvalue = getvalue

    def getvalue(self):
        """Get the string containing all of the written record.

        This function can also be called after the writer is closed.

        :returns: a string
        """
        return self._getvalue()

    def close(self):
        """Close the writer

        If the reader wasn't previously closed then close it. This will
        set the location properties to their final values, close any
        files that the writer may have opened, and set ``self.closed`` to False.
        
        ``self.getvalue()`` will still work after the file is closed.
        """
        if self.closed:
            return
        self.closed = True
        try:
            self._structure_writer.close()
        except StopIteration:
            pass

############################


# This is the public API for formats.

class Format(object):
    """Information about a toolkit format.

    Use :func:`chemfp.toolkit.get_format` and related functions to return
    a Format instance.

    The public properties are:

    .. py:attribute::toolkit_name

       the toolkit name; either "rdkit", "openeye", or "openbabel"

    .. py:attribute::name

       the format name, without any compression information

    .. py:attribute::compression

       the compression type: "" for uncompressed, "gz" for gzip

    .. py:attribute::record_format

       the normalized record format name. All SMILES formats are "smi",
       and this does not contain compression information

    """
    def __init__(self, toolkit_name, format_config, compression=None):
        """This constructor is not part of the public API"""
        self.toolkit_name = toolkit_name
        self._format_config = format_config
        self.name = format_config.register_name
        #self.record_format = format_config.record_format # XX I don't think I want this information here
        self.compression = compression  # "", "gz", "bz2", "xz" ...

    def __repr__(self):
        """Return a string like 'Format("openeye/sdf.gz")'"""
        if not self.compression:
            name = "%s/%s" % (self.toolkit_name, self.name)
        else:
            name = "%s/%s.%s" % (self.toolkit_name, self.name, self.compression)
        return "Format(%r)" % name

    @property
    def prefix(self):
        """Return the prefix to turn an unqualified parameter into a fully qualified parameter

        :returns: a string like "rdkit.smi" or "openbabel.sdf"
        """
        return self.toolkit_name + "." + self._format_config.register_name

    @property
    def is_input_format(self):
        """Return True if this toolkit can read molecules in this format"""
        return self._format_config.input is not None

    @property
    def is_output_format(self):
        """Return True if this toolkit can write molecules in this format"""
        return self._format_config.output is not None

    @property
    def is_available(self):
        """Return True if this version of the toolkit understands this format

        For example, if your version of RDKit does not support InChI then
        this would return False for the "inchi" and "inchikey" formats.
        """
        return self._format_config.is_available

    # I don't have a good name for this.

    # This is True for record-oriented formats (like "smi") that can
    # be stored directly into a file. It's False for the
    # non-record-oriented formats (like "smiles") where the output is
    # only part of some actual other format.

    @property
    def supports_io(self):
        """Return True if this format support reading or writing records

        This will return False for formats like "smistring" and "inchikeystring"
        because those are are not record-based formats.

        Note: I don't like this name. I may change it to ``is_record_format``.
        Let me know if you have ideas, or if changing the name will be a problem.
        """
        return self._format_config.supports_io

    @property
    def extensions(self):
        """Return a list of appropriate filename extensions for this format

        Returns an empty list if this format does not support io.
        """
        if not self._format_config.supports_io:
            return []
        extensions = [self.name]
        extensions.extend(self._format_config.alternate_extensions)
        return extensions

    # "settings" are dictionaries of key/value strings.
    # These can come from the command-line or environment variables.
    # These functions turn them into reader_args or writer_args.
    # Eg, turn {"rdkit.sdf.sanitize": "true"} or {"sdf.sanitize": "true"}
    # or {"sanitize": "true"} into {"sanitize": True}.

    # Qualified names include dots, like "rdkit.*.sanitize"
    # Unqualified names don't contain any dots.
    # Qualified names exist a configuration setting can support multiple toolkits.

    def _get_args_from_text_settings(self, settings, input_or_output):
        if input_or_output is None or not settings:
            return {}
        register_name = self._format_config.register_name

        # The search path is:
        #   rdkit.sdf.sanitize
        #   rdkit.*.sanitize
        #   sdf.sanitize
        #   sanitize
        # where the first match wins.
        
        prefix1 = self.toolkit_name + "." + register_name + "."
        prefix2 = self.toolkit_name + ".*."
        prefix3 = register_name + "."

        d = {}
        special = object()
        for arg_name, arg_parser in input_or_output.arg_info:
            name = prefix1 + arg_name
            value = settings.get(name, special)
            if value is special:
                name = prefix2 + arg_name
                value = settings.get(name, special)
                if value is special:
                    name = prefix3 + arg_name
                    value = settings.get(name, special)
                    if value is special:
                        name = arg_name
                        value = settings.get(name, special)
                        if value is special:
                            continue

            exc_obj = None
            try:
                parsed_value = arg_parser(value)
            except ValueError as err:
                exc_obj = ValueError("Unable to parse %s setting %s (%r): %s"
                                    % (self.toolkit_name, name, value, err))
            else:
                d[arg_name] = parsed_value
            if exc_obj is not None:
                # Prevent chained exceptions, in a Python 2.7 compatible way.
                raise exc_obj

        return d

    def get_reader_args_from_text_settings(self, reader_settings):
        """Process the *reader_settings* and return the *reader_args* for this format.

        This function exists to help convert string settings, eg, from the
        command-line or a configuration, into usable *reader_args*.

        Setting names may be fully-qualified names like "rdkit.sdf.sanitize",
        partially qualified names like "rdkit.*.sanitize" or "openeye.smi.delimiter",
        or unqualified names like "delimiter". The qualifiers act as a namespace
        so the settings can be specified without needing to know the actual
        toolkit or format.

        The function turns the format-appropriate qualified names into unqualified
        ones and converts the string values into usable Python objects. For example:

          >>> from chemfp import rdkit_toolkit  as T
          >>> fmt = T.get_format("smi")
          >>> fmt.get_reader_args_from_text_settings({"rdkit.*.sanitize": "true", "delimiter": "to-eol"})
          {'delimiter': 'to-eol', 'sanitize': True}

        :param reader_settings: the reader settings
        :type reader_settings: a dictionary with string keys and values
        :returns: a dictionary of unqualified argument names as keys and processed Python values as values
        """
        return self._get_args_from_text_settings(reader_settings, self._format_config.input)

    def get_writer_args_from_text_settings(self, writer_settings):
        """Process *writer_settings* and return the *writer_args* for this format.

        This function exists to help convert string settings, eg, from the
        command-line or a configuration, into usable *writer_args*.

        Setting names may be fully-qualified names like "rdkit.sdf.kekulize",
        partially qualified names like "rdkit.*.delimiter" or "openeye.smi.delimiter",
        or unqualified names like "delimiter". The qualifiers act as a namespace
        so the settings can be specified without needing to know the actual
        toolkit or format.

        The function turns the format-appropriate qualified names into unqualified
        ones and converts the string values into usable Python objects. For example:

          >>> from chemfp import rdkit_toolkit  as T
          >>> fmt = T.get_format("smi")
          >>> fmt.get_writer_args_from_text_settings({"rdkit.*.kekuleSmiles": "true", "canonical": "false"})
          {'kekuleSmiles': True, 'canonical': False}

        :param writer_settings: the writer settings
        :type writer_settings: a dictionary with string keys and values
        :returns: a dictionary of unqualified argument names as keys and processed Python values as values
        """
        return self._get_args_from_text_settings(writer_settings, self._format_config.output)


    # Turn reader_args/writer_args into settings
    # XXX TODO: I have nothing which does this inverse.
    ## def _get_settings_from_args(self, settings, qualified, input_or_output):
        
    ## def get_settings_from_reader_args(self, reader_args, qualified=True):
    ##     pass

    ## def get_settings_from_writer_args(self, writer_args, qualified=True):
    ##     pass

    # "Qualified" names have dots. "Unqualified" names don't.


    # Turn fully, partial, or unqualified names into unqualified names.
    def _get_unqualified_args(self, args, input_or_output):
        if input_or_output is None:
            return {}
        if not args:
            return input_or_output.defaults.copy()
        register_name = self._format_config.register_name

        # The search path is:
        #   rdkit.sdf.sanitize
        #   rdkit.*.sanitize
        #   sdf.sanitize
        #   sanitize
        
        prefix1 = self.toolkit_name + "." + register_name + "."
        prefix2 = self.toolkit_name + ".*."
        prefix3 = register_name + "."

        d = {}
        special = object()
        for arg_name, arg_parser in input_or_output.arg_info:
            value = args.get(prefix1 + arg_name, special)
            if value is special:
                value = args.get(prefix2 + arg_name, special)
                if value is special:
                    value = args.get(prefix3 + arg_name, special)
                    if value is special:
                        value = args.get(arg_name, special)
                        if value is special:
                            # Nothing found. Use the default
                            d[arg_name] = input_or_output.defaults[arg_name]
                            continue

            # Use the found value.
            d[arg_name] = value
            
        return d

    def get_default_reader_args(self):
        """Return a dictionary of the default reader arguments

        The keys are unqualified (ie, without dots).

          >>> from chemfp import openbabel_toolkit as T
          >>> fmt = T.get_format("smi")
          >>> fmt.get_default_reader_args()
          {'has_header': False, 'delimiter': None, 'options': None}

        :returns: a dictionary of string keys and Python objects for values
        """
        input = self._format_config.input
        if input is None:
            return {}
        return input.defaults.copy()

    def get_default_writer_args(self):
        """Return a dictionary of the default writer arguments

        The keys are unqualified (ie, without dots).

          >>> from chemfp import openbabel_toolkit as T
          >>> fmt = T.get_format("smi")
          >>> fmt.get_default_writer_args()
          {'explicit_hydrogens': False, 'isomeric': True, 'delimiter': None,
          'options': None, 'canonicalization': 'default'}

        :returns: a dictionary of string keys and Python objects for values
        """
        output = self._format_config.output
        if output is None:
            return {}
        return output.defaults.copy()

    def get_unqualified_reader_args(self, reader_args):
        """Convert possibly qualified reader args into unqualified reader args for this format

        The *reader_args* dictionary can be confusing because of the
        priority rules in how to resolve qualifiers, and because it
        can include irrelevant parameters, which are ignored.
        
        The get_unqualified_reader_args function applies the qualifier resolution
        algorithm and removes irrelevant parameters to return a dictionary
        containing the equivalent unqualified reader args dictionary for this format.

          >>> from chemfp import rdkit_toolkit as T
          >> fmt = T.get_format("smi")
          >>> fmt.get_unqualified_reader_args({"rdkit.*.delimiter": "tab", "smi.sanitize": False, "X": "Y"})
          {'delimiter': 'tab', 'has_header': False, 'sanitize': False}
          >>> fmt = T.get_format("can")
          >>> fmt.get_unqualified_reader_args({"rdkit.*.delimiter": "tab", "smi.sanitize": False, "X": "Y"})
          {'delimiter': 'tab', 'has_header': False, 'sanitize': True}

        :parameters reader_args: reader arguments, which can contain qualified and unqualified arguments
        :type reader_args: a dictionary with string keys and Python values
        :returns: a dictionary of reader arguments, containing only unqualified arguments
          appropriate for this format.
        """
        return self._get_unqualified_args(reader_args, self._format_config.input)

    def get_unqualified_writer_args(self, writer_args):
        """Convert possibly qualified writer args into unqualified writer args for this format

        The *writer_args* dictionary can be confusing because of the
        priority rules in how to resolve qualifiers, and because it
        can include irrelevant parameters, which are ignored.

        The get_unqualified_writer_args function applies the qualifier resolution
        algorithm and removes irrelevant parameters to return a dictionary
        containing the equivalent unqualified writer args dictionary for this format.

          >>> from chemfp import rdkit_toolkit as T
          >>> fmt = T.get_format("smi")
          >>> fmt.get_unqualified_writer_args({"rdkit.*.delimiter": "tab", "smi.kekuleSmiles": True, "X": "Y"})
          {'isomericSmiles': True, 'delimiter': 'tab', 'kekuleSmiles': True, 'allBondsExplicit': False, 'canonical': True}
          >>> fmt = T.get_format("can")
          >>> fmt.get_unqualified_writer_args({"rdkit.*.delimiter": "tab", "smi.kekuleSmiles": True, "X": "Y"})
          {'isomericSmiles': False, 'delimiter': 'tab', 'kekuleSmiles': False, 'allBondsExplicit': False, 'canonical': True}

        :parameters writer_args: writer arguments, which can contain qualified and unqualified arguments
        :type writer_args: a dictionary with string keys and Python values
        :returns: a dictionary of writer arguments, containing only unqualified arguments
          appropriate for this format.
        """

        return self._get_unqualified_args(writer_args, self._format_config.output)


# End of public API. The FormatRegistry will be public someday.

class FormatRegistry(object):
    def __init__(self, external_name, toolkit_name):
        self.external_name = external_name
        self.toolkit_name = toolkit_name

        self._format_config_by_name = {}
        self._alternate_extensions = {}

    def __contains__(self, format_name):
        return format_name in self._format_config_by_name

    def register_format(self, format_config):
        register_name = format_config.register_name
        assert register_name not in self._format_config_by_name, register_name
        self._format_config_by_name[register_name] = format_config

        if format_config.is_available:
            for alternate in format_config.alternate_extensions:
                assert alternate not in self._alternate_extensions, alternate
                self._alternate_extensions[alternate] = register_name

    def register_formats(self, *format_configs):
        for format_config in format_configs:
            self.register_format(format_config)


    def get_format_config(self, register_name):
        format_config = self._format_config_by_name.get(register_name, None)
        if format_config is None:
            raise ValueError("%s does not support the %r format"
                             % (self.external_name, register_name))
        if not format_config.is_available:
            raise ValueError("This version of the %s toolkit does not support the %r format"
                             % (self.external_name, register_name))
        return format_config

    def get_input_format_config(self, register_name):
        format_config = self.get_format_config(register_name)
        if format_config.input is None:
            raise ValueError("%s does not support %r as an input format"
                             % (self.external_name, register_name))
        return format_config

    def get_output_format_config(self, register_name):
        format_config = self.get_format_config(register_name)
        if format_config.output is None:
            raise ValueError("%s does not support %r as an output format"
                             % (self.external_name, register_name))
        return format_config

    ## def check_source(self, source=None, format=None):
    ##     pass

    ## def check_destination(self, destination=None, format=None):
    ##     pass
        

    # Get a dictionary of input/output formats. The compression is None

    def get_formats(self, include_unavailable=False):
        toolkit_name = self.toolkit_name
        return [Format(toolkit_name, format_config, None)
                  for (register_name, format_config) in self._format_config_by_name.items()
                    if (include_unavailable or format_config.is_available) ]

    def get_input_formats(self):
        toolkit_name = self.toolkit_name
        return [Format(toolkit_name, format_config, None)
                  for (register_name, format_config) in self._format_config_by_name.items()
                    if format_config.is_available and format_config.input is not None]

    def get_output_formats(self):
        toolkit_name = self.toolkit_name
        return [Format(toolkit_name, format_config, None)
                  for (register_name, format_config) in self._format_config_by_name.items()
                    if format_config.is_available and format_config.output is not None]

    # Get a Format given a format string (which can include compression)
    def _normalize_format(self, format):
        if format is None:
            return "smi", None
        if isinstance(format, _compat.filename_types):
            return io._normalize_format_name(format)
        return format.name, format.compression

    def get_format(self, format):
        register_name, compression = self._normalize_format(format)
        format_config = self.get_format_config(register_name)
        return Format(self.toolkit_name, format_config, compression)
    
    def get_input_format(self, format):
        register_name, compression = self._normalize_format(format)
        format_config = self.get_input_format_config(register_name)
        return Format(self.toolkit_name, format_config, compression)

    def get_output_format(self, format):
        register_name, compression = self._normalize_format(format)
        format_config = self.get_output_format_config(register_name)
        return Format(self.toolkit_name, format_config, compression)

    # Get a Format given a source/destination and optional format name

    def get_input_format_from_source(self, source=None, format=None, default=("smi", "")):
        if format is not None and not isinstance(format, _compat.filename_types):
            if isinstance(format, Format):
                return format
            # Someone passed in an object with "name" and "compression".
            # Convert it to the special Format, which has the _format_config
            register_name, compression = format.name, format.compression
            if compression is None:
                compression = ""
        else:
            register_name, compression = io.normalize_input_format(source, format, default=default)
            # Support various extensions
            register_name = self._alternate_extensions.get(register_name, register_name)

        format_config = self.get_input_format_config(register_name)
        return Format(self.toolkit_name, format_config, compression)


    def get_output_format_from_destination(self, destination=None, format=None, default=("smi", "")):
        if format is not None and not isinstance(format, _compat.filename_types):
            if isinstance(format, Format):
                return format
            # Someone passed in an object with "name" and "compression".
            # Convert it to the special Format, which has the _format_config
            register_name, compression = format.name, format.compression
            if compression is None:
                compression = ""
        else:
            register_name, compression = io.normalize_output_format(destination, format, default=default)
            # Support various extensions
            register_name = self._alternate_extensions.get(register_name, register_name)

        format_config = self.get_output_format_config(register_name)
        return Format(self.toolkit_name, format_config, compression)


###### End of the public API

###### Support classes for format arguments


class CharacterDelimiter(object):
    def __init__(self, delimiter):
        self.delimiter_string = delimiter
        self.delimiter_bytes = delimiter.encode("ascii")
        self.name = delimiter

    def split_two_terms_string(self, line):
        # NOTE: the caller must remove the terminal newline
        if not line:
            return None
        terms = line.split(self.delimiter_string, 2)
        if not terms:
            return None
        if len(terms) == 1:
            return terms[0], None
        return terms[0], terms[1]
    
    def split_two_terms_bytes(self, line):
        # NOTE: the caller must remove the terminal newline
        if not line:
            return None
        terms = line.split(self.delimiter_bytes, 2)
        if not terms:
            return None
        if len(terms) == 1:
            return terms[0], None
        return terms[0], terms[1]
        
    def format_two_terms(self, col1, col2):
        if col2 is None:
            return col1 + "\n"
        return col1 + self.delimiter_string + col2 + "\n"

    def format_two_terms_bytes(self, col1, col2):
        if col2 is None:
            return col1 + b"\n"
        return col1 + self.delimiter_bytes + col2 + b"\n"

    def get_format_string(self):
        if self.delimiter_string == "%":
            return "%s%%%s"
        else:
            return "%s" + self.delimiter_string + "%s\n"

class WhitespaceDelimiter(object):
    name = "whitespace"
    code = 1
    def split_two_terms_string(self, line):
        terms = line.split()
        if not terms:
            return None
        if len(terms) == 1:
            return terms[0], None
        return terms[0], terms[1]

    split_two_terms_bytes = split_two_terms_string

    def format_two_terms(self, col1, col2):
        if col2 is None:
            return col1 + "\n"
        return "%s %s\n" % (col1, col2)

    def format_two_terms_bytes(self, col1, col2):
        if col2 is None:
            return col1 + b"\n"
        return b"%s %s\n" % (col1, col2)

    def get_format_string(self):
        return "%s %s\n"

class ToEolDelimiter(object):
    name = "to-eol"
    code = 0
    def split_two_terms_string(self, line):
        terms = line.split(None, 1)
        if not terms:
            return None
        if len(terms) == 1:
            return terms[0], None
        return terms[0], terms[1]

    split_two_terms_bytes = split_two_terms_string

    def format_two_terms(self, col1, col2):
        if col2 is None:
            return col1 + "\n"
        return "%s %s\n" % (col1, col2)
    
    def format_two_terms_bytes(self, col1, col2):
        if col2 is None:
            return col1 + b"\n"
        return b"%s %s\n" % (col1, col2)

    def get_format_string(self):
        return "%s %s\n"

_delimiters = {
    "space": CharacterDelimiter(u" "),
    " ": CharacterDelimiter(u" "),
    "tab": CharacterDelimiter(u"\t"),
    "\t": CharacterDelimiter(u"\t"),
    "whitespace": WhitespaceDelimiter(),
    None: ToEolDelimiter(),
    "to-eol": ToEolDelimiter(),
    "native": ToEolDelimiter(),  # Not really supposed to need this one
    }

## def get_delimiter_simple(delimiter):
##     try:
##         return _delimiters[delimiter]
##     except KeyError:
##         pass
##     raise ValueError("Unsupported delimiter %r" % (delimiter,))
    

def get_delimiter(format_config, reader_args, default=None):
    delimiter = reader_args["delimiter"]
    if delimiter == "native":
        delimiter = default
    try:
        return _delimiters[delimiter]
    except KeyError:
        pass
    
    raise ValueError("Unsupported delimiter %r for  %r"
                     % (delimiter, format_config.register_name))
        


def parse_boolean(value):
    if value in ("1", "True", "true"):
        return True
    if value in ("0", "False", "false"):
        return False
    raise ValueError("Unsupported boolean value")

## def parse_delimiter(value):
##     if value in _delimiters:
##         return value
##     else:
##         raise ValueError("Unsupported delimiter")

class DelimiterParser(object):
    def __init__(self, default=None):
        self.default = default
    def __call__(self, value):
        if value == "native":
            value = self.default
        if value in _delimiters:
            return value
        raise ValueError("Unsupported delimiter")

parse_delimiter = DelimiterParser("to-eol")

###### Base toolkit type

def only_utf8(format, encoding, encoding_errors):
    if encoding != "utf8":
        raise EncodingError("The %s format only accepts 'utf8' encoding, not %r"
                            % (format.prefix, encoding))
    if encoding_errors != "strict":
        raise EncodingError("The %s format only accepts 'strict' encoding errors, not %r"
                            % (format.prefix, encoding_errors))

class BaseToolkit(object):
    def __init__(self, format_registry):
        self.format_registry = format_registry


    # Implementations

    # Process the inputs, check with the formats, etc. and pass the
    # processed options to the "*_impl" sub-methods. This lets the
    # specific toolkits handle the actual back-end details.

    def read_molecules(self,
                       source=None, format=None, id_tag=None, reader_args=None,
                       errors="strict", location=None,
                       encoding="utf8", encoding_errors="strict"):
        # Figure out the actual format type
        format = self.format_registry.get_input_format_from_source(source, format)
        format_config = self.format_registry.get_input_format_config(format.name)
        compression = format.compression

        if not format.is_input_format:
            raise TypeError("%s %r is not an input format"
                            % (self.format_registry.external_name, format.name))

        if not format.supports_io:
            raise TypeError(
                "%s format %r does not support file I/O; try parse_molecule() instead"
                % (self.format_registry.external_name, format.name))

        self._check_read_encoding(format, encoding, encoding_errors)

        # Convert qualified arguments into unqualified ones
        reader_args = format.get_unqualified_reader_args(reader_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        # Track location
        if location is None:
            location = io.Location.from_source(source)
        location.save(record_format=format_config.record_format)

        # Pass to the back-end toolkit
        return self._read_molecules_impl(format_config, source, compression, id_tag,
                                         reader_args, error_handler, location, encoding, encoding_errors)

    def _check_read_encoding(self, format, encoding, encoding_errors):
        only_utf8(format, encoding, encoding_errors)
    
    def _read_molecules_impl(self, format_config, source, compression, id_tag,
                             reader_args, error_handler, location, encoding, encoding_errors):
        # The default implementation uses _read_ids_and_molecules_impl
        reader = self._read_ids_and_molecules_impl(format_config, source, compression, id_tag,
                                                   reader_args, error_handler, location, encoding, encoding_errors)
        return MoleculeReader(reader.metadata,
                              (mol for (id, mol) in reader),
                              reader.location)
    
    def read_ids_and_molecules(self,
                               source=None, format=None, id_tag=None, reader_args=None,
                               errors="strict", location=None,
                               encoding="utf8", encoding_errors="strict"):
        # Figure out the actual format type
        format = self.format_registry.get_input_format_from_source(source, format)
        format_config = self.format_registry.get_input_format_config(format.name)
        compression = format.compression

        if not format.is_input_format:
            raise TypeError("%s %r is not an input format"
                            % (self.format_registry.external_name, format.name))

        if not format.supports_io:
            raise TypeError(
                "%s format %r does not support file I/O; try parse_id_and_molecule() instead"
                % (self.format_registry.external_name, format.name))
        
        self._check_read_encoding(format, encoding, encoding_errors)

        # Convert qualified arguments into unqualified ones
        reader_args = format.get_unqualified_reader_args(reader_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        # Track location
        if location is None:
            location = io.Location.from_source(source)
        location.save(record_format=format_config.record_format)

        # Pass to the back-end toolkit
        return self._read_ids_and_molecules_impl(format_config, source, compression, id_tag,
                                                 reader_args, error_handler, location, encoding, encoding_errors)

    def _read_ids_and_molecules_impl(self, format_config, source, compression, id_tag,
                                     reader_args, error_handler, location, encoding, encoding_errors):
        return format_config.read_ids_and_molecules(source, compression, id_tag,
                                                    reader_args, error_handler, location,
                                                    encoding, encoding_errors)


    def read_molecules_from_string(self,
                                   content, format, id_tag=None, reader_args=None,
                                   errors="strict", location=None):
        if format is None:
            raise ValueError("Missing format, and unable to guess structure format based on content string")

        # Figure out the actual format type
        format = self.format_registry.get_input_format(format)
        format_config = self.format_registry.get_input_format_config(format.name)
        compression = format.compression

        if not format.is_input_format:
            raise TypeError("%s %r is not an input format"
                            % (self.format_registry.external_name, format.name))

        if not format.supports_io:
            raise TypeError(
                "%s format %r does not support file I/O; try parse_molecule() instead"
                % (self.format_registry.external_name, format.name))

        # Convert qualified arguments into unqualified ones
        reader_args = format.get_unqualified_reader_args(reader_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        # Track location
        if location is None:
            location = io.Location("<string>")
        location.save(record_format=format_config.record_format)

        # Pass to the back-end toolkit
        return self._read_molecules_from_string_impl(format_config, content, compression, id_tag,
                                                     reader_args, error_handler, location)

    def _read_molecules_from_string_impl(self, format_config, content, compression, id_tag,
                                         reader_args, error_handler, location):
        # The default implementation uses _read_ids_and_molecules_from_string_impl
        reader = self._read_ids_and_molecules_from_string_impl(format_config, content, compression, id_tag,
                                                               reader_args, error_handler, location)
        return MoleculeReader(reader.metadata,
                              (mol for (id, mol) in reader),
                              reader.location)
        

    def read_ids_and_molecules_from_string(self,
                                           content, format, id_tag=None, reader_args=None,
                                           errors="strict", location=None):
        if format is None:
            raise ValueError("Missing format, and unable to guess structure format based on content string")

        # Figure out the actual format type
        format = self.format_registry.get_input_format(format)
        format_config = self.format_registry.get_input_format_config(format.name)
        compression = format.compression

        if not format.is_input_format:
            raise TypeError("%s %r is not an input format"
                            % (self.format_registry.external_name, format.name))

        if not format.supports_io:
            raise TypeError(
                "%s format %r does not support file I/O; try parse_id_and_molecule() instead"
                % (self.format_registry.external_name, format.name))

        # Convert qualified arguments into unqualified ones
        reader_args = format.get_unqualified_reader_args(reader_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        # Track location
        if location is None:
            location = io.Location("<string>")
        location.save(record_format=format_config.record_format)

        # Pass to the back-end toolkit
        return self._read_ids_and_molecules_from_string_impl(format_config, content, compression, id_tag,
                                                             reader_args, error_handler, location)

    def _read_ids_and_molecules_from_string_impl(self, format_config, content, compression, id_tag,
                                                 reader_args, error_handler, location):
        return format_config.read_ids_and_molecules_from_string(content, compression, id_tag,
                                                                reader_args, error_handler, location)


    def make_id_and_molecule_parser(self, format, id_tag=None, reader_args=None, errors="strict"):
        if format is None:
            raise ValueError("Missing required 'format'")
        
        # Figure out the actual format type
        format = self.format_registry.get_input_format(format)
        format_config = self.format_registry.get_input_format_config(format.name)
        compression = format.compression
        
        # Convert qualified arguments into unqualified ones
        reader_args = format.get_unqualified_reader_args(reader_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        # Pass to the back-end toolkit
        return self._make_id_and_molecule_parser_impl(format_config, compression,
                                                      id_tag, reader_args, error_handler)

    def _make_id_and_molecule_parser_impl(self, format_config, compression, id_tag, reader_args, error_handler):
        parser = format_config.make_id_and_molecule_parser(id_tag, reader_args, error_handler)
        if compression in (None, ""):
            return parser
        if compression == "gz":
            from gzip import GzipFile
            
            def gzip_decompressor(content):
                infile = GzipFile(fileobj=_compat.BytesIO(content))
                return parser(infile.read())

            gzip_decompressor.__name__ = parser.__name__ + "_gz"
            return gzip_decompressor

        if compression == "zst":
            try:
                import zstandard as zstd
            except ImportError as err:
                raise ValueError("Please install the zstandard module for zst compression support: %s"
                                     % (err,))

            dctx = zstd.ZstdDecompressor()
            def zst_decompressor(content):
                bio = _compat.BytesIO(content)
                return parser(dctx.stream_reader(bio).read())

            zst_decompressor.__name__ = parser.__name__ + "_zst"
            return zst_decompressor
        
        if compression == "zlib":
            import zlib
            def zlib_decompressor(content):
                return parser(zlib.decompress(content))

            zlib_decompressor.__name__ = parser.__name__ + "_zlib"
            return zlib_decompressor

        raise ValueError("%s does not support compression type %r"
                         % (self.format_registry.external_name, compression))
        

    def parse_molecule(self, content, format, id_tag=None, reader_args=None, errors="strict"):
        if format is None:
            raise ValueError("Missing 'format' and unable to guess structure format based on content string")
        
        # Figure out the actual format type
        format = self.format_registry.get_input_format(format)
        format_config = self.format_registry.get_input_format_config(format.name)
        compression = format.compression

        # Convert qualified arguments into unqualified ones
        reader_args = format.get_unqualified_reader_args(reader_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        # Decompress here. The underlying formats only take uncompressed content.
        # TODO: If decompression fails, should I use the error_handler?
        content = io._decompress_string(content, compression, self.format_registry.external_name)

        # Pass to the back-end toolkit
        return self._parse_molecule_impl(format_config, content,
                                         id_tag, reader_args, error_handler)

    def _parse_molecule_impl(self, format_config, content, id_tag, reader_args, error_handler):
        # Use the parse_id_and_molecule implementation, then throw away the id.
        id, mol = format_config.parse_id_and_molecule(content, id_tag, reader_args, error_handler)
        return mol


    def parse_id_and_molecule(self, content, format, id_tag=None, reader_args=None, errors="strict"):
        if format is None:
            raise ValueError("Missing 'format' and unable to guess structure format based on content string")
        
        # Figure out the actual format type
        format = self.format_registry.get_input_format(format)
        format_config = self.format_registry.get_input_format_config(format.name)
        
        # Decompress here. The underlying formats only take uncompressed content.
        # TODO: If decompression fails, should I use the error_handler?
        content = io._decompress_string(content, format.compression, self.format_registry.external_name)
        
        # Convert qualified arguments into unqualified ones
        reader_args = format.get_unqualified_reader_args(reader_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        # Pass to the back-end toolkit
        return self._parse_id_and_molecule_impl(format_config, content,
                                                id_tag, reader_args, error_handler)
        
    def _parse_id_and_molecule_impl(self, format_config, content,
                                    id_tag, reader_args, error_handler):
        return format_config.parse_id_and_molecule(content, id_tag, reader_args, error_handler)

    ##### Output

    def _check_write_encoding(self, format, encoding, encoding_errors):
        only_utf8(format, encoding, encoding_errors)

    def create_string(self, mol, format, id=None, writer_args=None, errors="strict"):
        if format is None:
            raise ValueError("create_string requires a format, not None")

        format = self.format_registry.get_output_format(format)
        format_config = self.format_registry.get_format_config(format.name)
        compression = format.compression

        if compression not in (None, ""):
            raise ValueError("create_string() does not support compression. Use create_bytes()")
        
        # Convert qualified arguments into unqualified ones
        writer_args = format.get_unqualified_writer_args(writer_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        return self._create_string_impl(format_config, mol, id, writer_args, error_handler)

    def _create_string_impl(self, format_config, mol, id, writer_args, error_handler):
        return format_config.create_string(mol, id, writer_args, error_handler)
    
    def create_bytes(self, mol, format, id=None, writer_args=None, errors="strict", level=None):
        if format is None:
            raise ValueError("create_bytes requires a format, not None")
        
        format = self.format_registry.get_output_format(format)
        format_config = self.format_registry.get_format_config(format.name)
        compression = format.compression

        if compression not in (None, "", "gz", "zlib", "zst"):
            raise ValueError("create_bytes() does not support compressed format %r" % (format,))
        elif compression == "":
            compression = None

        # Convert qualified arguments into unqualified ones
        writer_args = format.get_unqualified_writer_args(writer_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        return self._create_bytes_impl(format_config, compression, mol, id, writer_args, error_handler, level)

    
    def _create_bytes_impl(self, format_config, compression, mol, id, writer_args, error_handler, level):
        s = format_config.create_bytes(mol, id, writer_args, error_handler)
        return io._compress_string(s, compression, "create_bytes", level)

    def open_molecule_writer(self,
                             destination=None, format=None, writer_args=None,
                             errors="strict", location=None,
                             encoding="utf8", encoding_errors="strict",
                             level=None):
        # Figure out the actual format type
        format = self.format_registry.get_output_format_from_destination(destination, format)
        format_config = self.format_registry.get_format_config(format.name)
        compression = format.compression

        if not format.is_output_format:
            raise TypeError("%s %r is not an output format"
                            % (self.format_registry.external_name, format.name))

        if not format.supports_io:
            raise TypeError(
                "%s format %r does not support file I/O; try create_string instead"
                % (self.format_registry.external_name, format.name))
        
        self._check_write_encoding(format, encoding, encoding_errors)
        
        # Convert qualified arguments into unqualified ones
        writer_args = format.get_unqualified_writer_args(writer_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        # Track location
        if location is None:
            location = io.Location.from_destination(destination)
        location.save(record_format=format_config.record_format)

        # Pass to the back-end toolkit
        return self._open_molecule_writer_impl(format_config, destination, compression,
                                               writer_args, error_handler, location,
                                               encoding, encoding_errors, level)

    def _open_molecule_writer_impl(self, format_config, destination, compression,
                                   writer_args, error_handler, location,
                                   encoding, encoding_errors, level):
        return format_config.open_molecule_writer(destination, writer_args,
                                                  error_handler, location,
                                                  encoding, encoding_errors, level)

    def open_molecule_writer_to_bytes(self, format, writer_args=None,
                                      errors="strict", location=None, level=None):
                                      
        if format is None:
            raise ValueError("open_molecule_writer_to_bytes requires a format, not None")

        format = self.format_registry.get_output_format(format)
        format_config = self.format_registry.get_format_config(format.name)
        compression = format.compression

        if not format.is_output_format:
            raise TypeError("%s %r is not an output format"
                            % (self.format_registry.external_name, format.name))

        if not format.supports_io:
            raise TypeError(
                "%s format %r does not support file I/O; try create_bytes instead"
                % (self.format_registry.external_name, format.name))

        # Convert qualified arguments into unqualified ones
        writer_args = format.get_unqualified_writer_args(writer_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        # Track location
        if location is None:
            location = io.Location("<bytes>")
        location.save(record_format=format_config.record_format)

        # Pass to the back-end toolkit
        return self._open_molecule_writer_to_bytes_impl(format_config, compression,
                                                        writer_args, error_handler, location, level)

    def _open_molecule_writer_to_bytes_impl(self, format_config, compression,
                                             writer_args, error_handler, location, level):
        return format_config.open_molecule_writer_to_bytes(compression, writer_args, error_handler, location, level)

    def open_molecule_writer_to_string(self,
                                       format, writer_args=None, errors="strict", location=None):
        if format is None:
            raise ValueError("open_molecule_writer_to_string requires a format, not None")

        format = self.format_registry.get_output_format(format)
        format_config = self.format_registry.get_format_config(format.name)
        compression = format.compression
        
        if compression not in (None, ""):
            raise ValueError("open_molecule_writer_to_string() does not support compression. "
                             "Use open_molecule_writer_to_bytes()")

        if not format.is_output_format:
            raise TypeError("%s %r is not an output format"
                            % (self.format_registry.external_name, format.name))

        if not format.supports_io:
            raise TypeError(
                "%s format %r does not support file I/O; try create_string instead"
                % (self.format_registry.external_name, format.name))
            
        # Convert qualified arguments into unqualified ones
        writer_args = format.get_unqualified_writer_args(writer_args)

        # Get the correct error handler
        error_handler = io.get_parse_error_handler(errors)

        # Track location
        if location is None:
            location = io.Location("<string>")
        location.save(record_format=format_config.record_format)

        # Pass to the back-end toolkit
        return self._open_molecule_writer_to_string_impl(format_config, 
                                                         writer_args, error_handler, location)

    def _open_molecule_writer_to_string_impl(self, format_config, 
                                             writer_args, error_handler, location):
        return format_config.open_molecule_writer_to_string(writer_args, error_handler, location)


class NotIoFormatConfigMixin(object):
    supports_io = False
    read_ids_and_molecules = None
    read_ids_and_molecules_from_string = None
    open_molecule_writer = None
    open_molecule_writer_to_string = None
