from __future__ import absolute_import, print_function

# Copyright (c) 2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##


import sys
try:
    import ctypes
    from ctypes import cdll
except ImportError:
    # Could happen if this Python wasn't compiled with ctypes
    class ctypes:
        c_char_p = c_void_p = c_int = c_uint = None
        @staticmethod
        def POINTER(x):
            return None
        
    class cdll:
        @staticmethod
        def LoadLibrary(s):
            return None
    
import os
import gzip

# My timings show that I can read faster if I do the gzip reading myself.
# The writer isn't faster

BadGzipFile = getattr(gzip, "BadGzipFile", OSError)

Z_OK = 0
Z_STREAM_END = 1
Z_NEED_DICT = 2
Z_ERRNO = -1
Z_STREAM_ERROR = -2
Z_DATA_ERROR = -3
Z_MEM_ERROR = -4
Z_BUF_ERROR = -5
Z_VERSION_ERROR = -6

# Let people override my code, in case it's broken.

libz = None
if ctypes.c_char_p is None:
    NO_ZLIB_ERROR_MSG = "could not import ctypes"
elif os.environ.get("CHEMFP_USE_SYSTEM_GZIP", "0") == "1":
    NO_ZLIB_ERROR_MSG = "disabled by environment variable"
else:
    if sys.platform == "darwin":
        libz_libraries = ["libz.dylib"]
    else:
        libz_libraries = ["libz.so.1", "libz.so"]

    for libz_library in libz_libraries:
        try:
            libz = cdll.LoadLibrary(libz_library)
        except OSError:
            pass
        else:
            # Make sure it looks right
            try:
                libz.gzopen
            except AttributeError:
                libz = None
        if libz is not None:
            break
    NO_ZLIB_ERROR_MSG = "could not load zlib"

#libz = None
is_available = (libz is not None)

def get_fd(stream):
    if not is_available:
        return None
    func = getattr(stream, "fileno", None)
    if func is None:
        return None
    if not callable(func):
        return None
    import io
    try:
        return func()
    except io.UnsupportedOperation:
        return None


if libz is None:
    class not_implemented(object):
        def __init__(self, *args):
            raise NotImplementedError(NO_ZLIB_ERROR_MSG)
        
    # Let the ctype declarations work
    class libz:
        class gzopen(not_implemented): pass
        class gzdopen(not_implemented): pass
        class gzbuffer(not_implemented): pass
        class gzdirect(not_implemented): pass
        class gzread(not_implemented): pass
        class gzwrite(not_implemented): pass
        class gzeof(not_implemented): pass
        class gzerror(not_implemented): pass
        class gzclose(not_implemented): pass

_gzopen = libz.gzopen
# (const char *path, const char *mode)
_gzopen.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
_gzopen.restype = ctypes.c_void_p

_gzdopen = libz.gzdopen
# (int fd, const char *mode)
_gzdopen.argtypes = [ctypes.c_int, ctypes.c_char_p]
_gzdopen.restype = ctypes.c_void_p
try:
    _gzbuffer = libz.gzbuffer
except AttributeError:
    # This was added in 1.2.3.5 (8 Jan 2010)
    # chemfp 3.4 supports Python's manylinux2010
    # The libz in that release doesn't implement gzbuffer
    def _gzbuffer(fd, size):
        return 0
else:
    # (gzFile file, unsigned size)
    _gzbuffer.argtypes = [ctypes.c_void_p, ctypes.c_uint]
    # 0 on success, -1 on error
    _gzbuffer.restype = ctypes.c_int

_gzdirect = libz.gzdirect
# (gzFile file)
_gzdirect.argtypes = [ctypes.c_void_p]
# 1 for copied (uncompressed), 0 for gzip (decompressed)
_gzdirect.restype = ctypes.c_int

_gzread = libz.gzread
# (gzFile file, voidp buf, unsigned len)
_gzread.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_uint]
# number of uncompressed bytes actually read, less than
# len for end of file, or -1 for error.
_gzread.restype = ctypes.c_int

_gzwrite = libz.gzwrite
# (gzFile file, voidpc buf, unsigned len)
_gzwrite.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_uint]
# the number of uncompressed bytes written or 0 in case of error.
_gzwrite.restype = ctypes.c_int

_gzeof = libz.gzeof
# (gzFile file)
_gzeof.argtypes = [ctypes.c_void_p]
# 1 if the end-of-file indicator has been set while reading, else 0
_gzeof.restype = ctypes.c_int

_gzerror = libz.gzerror
# (gzFile file, int *errnum)
_gzerror.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_int)]
# returns error message, and sets errnum
_gzerror.restype = ctypes.c_char_p

_gzclose = libz.gzclose
# (gzFile file)
_gzclose.argtypes = [ctypes.c_void_p]
# zlib.h says: "Z_STREAM_ERROR if file is not valid, Z_ERRNO on a file
# operation error, Z_MEM_ERROR if out of memory, Z_BUF_ERROR if the
# last read ended in the middle of a gzip stream, or Z_OK on success."
_gzclose.restype = ctypes.c_int

####

def _name_str(s, name):
    if name is None:
        return s
    return s + ": " + name

def _get_zlib_exception_from_gzclose_errcode(errcode, name):
    if errcode == Z_OK:
        raise AssertionError("There's no exception to raise")
    
    if errcode == Z_BUF_ERROR:
        return EOFError(_name_str(
            "Compressed file ended before the "
            "end-of-stream marker was reached",
            name))
        
    #print("errno", errno)

    # zlib.h says: gzclose will return Z_STREAM_ERROR if file is not
    # valid, Z_ERRNO on a file operation error, Z_MEM_ERROR if out of
    # memory, Z_BUF_ERROR if the last read ended in the middle of a
    # gzip stream, or Z_OK on success

    if errcode == Z_STREAM_ERROR:
        import errno # I don't have any way to get the actual error code
        msg = "zlib gzcoose reports the file is not valid"
        if name is not None:
            msg += ": " + repr(name)
        raise OSError(errno.EBADF, msg)
        
    if errcode == Z_ERRNO:
        import errno # I don't have any way to get the actual error code
        msg = "zlib gzclose reports there was a system error"
        if name is not None:
            msg += ": " + repr(name)
        raise OSError(errno.EBADF, msg)
        
    if errcode == Z_MEM_ERROR:
        msg = "zlib ran out of memory"
        if name is not None:
            msg += ": " + repr(name)
        raise MemoryError(msg)
    
    raise AssertionError(("unknown zlib close error", errcode, name))
        
def _get_zlib_exception_from_gzerr(err, errstr_p, name):
    errcode = err.value
    if errcode == Z_OK:
        raise AssertionError("There's no exception to raise")

    if errcode == Z_STREAM_END:
        raise AssertionError("There's no exception to raise")

    if errcode == Z_BUF_ERROR:
        return EOFError(_name_str(
                "Compressed file ended before the "
                "end-of-stream marker was reached",
                name))

    errbytes = ctypes.string_at(errstr_p)
    if name is not None:
        i = errbytes.rfind(b": ")
        if i != -1:
            errstr = errbytes[i+2:].decode("ascii", errors="replace")
        else:
            errstr = errbytes.decode("utf8", errors="replace")
    else:
        errstr = errbytes.decode("utf8", errors="replace")


    msg = "zlib error %d (%r)" % (errcode, errstr)
    if name is not None:
        msg += ": " + repr(name)
    return BadGzipFile(msg)


def gzdopen_rb(fd, name=None, dup=True):
    if fd < 0:
        raise ValueError("file descriptor must be non-negative: %d" % (fd,))
    if dup:
        fd = os.dup(fd)
    gzf = _gzdopen(fd, b"rb")
    if not gzf:
        raise RuntimeError("Cannot create a gzip file object - out of memory?")

    return _gzopen_read_finish(gzf, name)

# Python 2.7 doesn't have os.fsencode()
# It doesnt need it either.
_fsencode = getattr(os, "fsencode", None)
if _fsencode is None:
    def _fsencode(s):
        return s

def gzopen_rb(path):
    # XXX what about Python 2.7?
    path_bytes = _fsencode(path)
    gzf = _gzopen(path_bytes, b"rb")
    if not gzf:
        with open(path, "rb"):
            pass
        import errno
        raise OSError(errno.EBADF, "zlib gzopen failed but Python succeeded?")

    return _gzopen_read_finish(gzf, path)

def _gzopen_read_finish(gzf, name):
    # zlib.h says: "A larger buffer size of, for example, 64K or 128K bytes
    # will noticeably increase the speed of decompression (reading)."
    retval = _gzbuffer(gzf, 128*1024)
    if retval != 0:
        _gzclose(gzf)
        raise RuntimeError("Cannot gzbuffer() the gzip file object - which should be impossible?")

    if _gzdirect(gzf) != 0:
        _gzclose(gzf)
        if name is not None:
            raise BadGzipFile("Not a gzipped file: %r" % (name,))
        else:
            raise BadGzipFile("Not a gzipped file")
    
    return GZFileReader(gzf, name)


def _get_compresslevel(compresslevel):
    if compresslevel is None:
        compresslevel = 9
    elif isinstance(compresslevel, int):
        if compresslevel < 1:
            compresslevel = 1
        if compresslevel > 9:
            compresslevel = 9
    else:
        raise ValueError("compresslevel must be an integer: %r" % (compresslevel,))

    return compresslevel

def _get_compresslevel_bytes(compresslevel):
    return str(_get_compresslevel(compresslevel)).encode("ascii")

def gzdopen_wb(fd, compresslevel=9, name=None, dup=True):
    if fd < 0:
        raise ValueError("file descriptor must be non-negative: %d" % (fd,))
    level = _get_compresslevel_bytes(compresslevel)
    if dup:
        fd = os.dup(fd)
    gzf = _gzdopen(fd, b"wb" + level)
    if not gzf:
        raise RuntimeError("Cannot create a gzip file object - out of memory?")

    return _gzopen_write_finish(gzf, name)

def gzopen_wb(path, compresslevel=9):
    # XXX what about Python 2.7?
    path_bytes = _fsencode(path)
    level = _get_compresslevel_bytes(compresslevel)
    #print("Open", b"wb" + level)
    gzf = _gzopen(path_bytes, b"wb" + level)
    if not gzf:
        raise AssertionError("better error handling here")
    return _gzopen_write_finish(gzf, path)

def _gzopen_write_finish(gzf, name):
    # zlib.h says: "A larger buffer size of, for example, 64K or 128K bytes
    # will noticeably increase the speed of decompression (reading)."
    retval = _gzbuffer(gzf, 128*1024)
    if retval != 0:
        _gzclose(gzf)
        raise RuntimeError("Cannot gzbuffer() the gzip file object - which should be impossible?")

    ## if _gzdirect(gzf) != 0:
    ##     _gzclose(gzf)
    ##     raise BadGzipFile("Not a gzipped file")
    
    return GZFileWriter(gzf, name)




class BaseGZFile(object):
    def __init__(self, gzf, name):
        self._gzf = gzf
        self.name = name
        self._buf = None
        self._buf_size = -1

    @property
    def closed(self):
        return self._gzf is None
    
    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __del__(self):
        self.close()

        
    def close(self):
        gzf = self._gzf
        if gzf is None:
            return
        self._gzf = None
        errcode = _gzclose(gzf)
        if errcode == Z_OK:
            pass
        elif errcode == Z_STREAM_END:
            # At the end of input
            pass
        elif errcode == Z_BUF_ERROR:
            # Like Python, I'll ignore incomplete gzip files at this point
            pass
        else:
            # I wonder what this is
            raise _get_zlib_exception_from_gzclose_errcode(errcode, self.name)

    def read(self, size=-1):
        import errno
        raise OSError(errno.EBADF, "read() on write-only GzipFile object")

    def write(self, data):
        import errno
        raise OSError(errno.EBADF, "write() on read-only GzipFile object")
    

class GZFileReader(BaseGZFile):
    _it = None
    def __iter__(self):
        it = self._it
        if it is None:
            self._it = it = self._get_iter()
        return it
            
    def _get_iter(self):
        prev = b""
        while 1:
            s = self.read(100000)
            if not s:
                if prev:
                    yield prev
                break
            s = prev + s
            lines = s.splitlines(True)
            if lines[-1][-1:] == "\n":
                prev = b""
                # 'yield from lines' doesn't work with Python 2.7
                for line in lines:
                    yield line
            else:
                prev = lines[-1]
                # 'yield from lines[:-1]' doesn't work with Python 2.7
                for line in lines[:-1]:
                    yield line
            
    def read(self, size=-1):
        if size is None or size < 0:
            blocks = []
            while 1:
                s = self.read(100000)
                if not s:
                    break
                blocks.append(s)
            return b"".join(blocks)
        
        gzf = self._gzf
        if gzf is None:
            raise ValueError("write to closed file")

        if size == 0:
            return b""
        
        buf_size = self._buf_size
        if buf_size < size:
            self._buf = buf = (ctypes.c_char * size)()
            self._buf_size = size
        else:
            buf = self._buf

        read_size = _gzread(gzf, buf, size)
        #print("read_size:", read_size)
        if read_size == size:
            return ctypes.string_at(buf, read_size)
        
        if read_size == -1:
            err = ctypes.c_int()
            errstr_p = _gzerror(gzf, err)
            raise _get_zlib_exception_from_gzerr(err, errstr_p, self.name)
        
        err = ctypes.c_int()
        errstr_p = _gzerror(gzf, err)
        #print("errvalue:", err.value, read_size)
        if err.value == Z_OK or err.value == Z_STREAM_END:
            if read_size == 0:
                return b""
            return ctypes.string_at(buf, read_size)
        
        raise _get_zlib_exception_from_gzerr(err, errstr_p, self.name)

class GZFileWriter(BaseGZFile):
    def write(self, data):        
        if not isinstance(data, bytes):
            raise TypeError("a bytes-like object is required, not %r" % (
                data.__class__.__name))
        
        gzf = self._gzf
        if gzf is None:
            raise ValueError("write to closed file")
        
        if not data:
            return 0

        n = len(data)
        num_written = _gzwrite(gzf, data, n)
        if num_written == 0:
            raise AssertionError
        if num_written != n:
            raise AssertionError
        return num_written
    
def read_raw_main(filename):
    if filename is None:
        return _read_raw_main(0)

    with open(filename, "rb") as f:
        return _read_raw_main(f.fileno())
        
def _read_raw_main(fd):
    N = 100000
    buf = (ctypes.c_char * N)()
    err = ctypes.c_int(0)

    gzf = _gzdopen(fd, b"rb")
    #print("gzdopen:", gzf)
    #print("gzbuffer:", _gzbuffer(gzf, 128*1024))
    _gzbuffer(gzf, 128*1024)
    #print("gzdirect:", _gzdirect(gzf))
    _gzdirect(gzf)
    
    total_size = 0
    while 1:
        read_size = _gzread(gzf, buf, N)
        #print("read:", read_size)
        if read_size == -1:
            raise AssertionError("cannot gzread")
        total_size += read_size
        if read_size < N:
            errstr_p = _gzerror(gzf, err)
            if err.value == Z_OK:
                #print("Stream end")
                break
            if err.value == Z_STREAM_END:
                break
            msg = ctypes.string_at(errstr_p).decode("utf8")
            msg = msg.replace("<fd:0>:", "<stdin>")
            print("Error?", errstr_p, err.value)
            print("msg:", msg)
            1/0

    return total_size

def read_gzio_main(filename):
    if filename is None:
        f = gzdopen_rb(0)
    else:
        f = gzopen_rb(filename)
    with f:
        n = 0
        while 1:
            s = f.read(1000000)
            if not s:
                break
            n += len(s)
        return n

def read_gzip_main(filename):
    if filename is None:
        f = gzip.open("/dev/stdin", "rb")
    else:
        f = gzip.open(filename, "rb")
    
    try:
        n = 0
        while 1:
            s = f.read(1000000)
            if not s:
                break
            n += len(s)
        return n
    finally:
        f.close()

### Writer

def get_write_chunks(
        filename,
        num_chunks=1):
    with open(filename, "rb") as f:
        content = f.read()
    num_bytes = len(content) // num_chunks
    chunks = []
    start = 0
    for i in range(1, num_chunks):
        end = start + num_bytes
        chunks.append(content[start:end])
        start = end
    chunks.append(content[start:])
    return chunks
    
def write_api_main(destination, compresslevel, chunks):
    if destination is None:
        f = gzdopen_wb(sys.stdout.fileno(), name="<stdout>", compresslevel=compresslevel)
    else:
        f = gzopen_wb(destination, compresslevel=compresslevel)
    with f:
        for chunk in chunks:
            #print("write api", file=sys.stderr)
            f.write(chunk)

def write_gzip_main(destination, compresslevel, chunks):
    if destination is None:
        f = gzip.open(sys.stdout.buffer, "wb", compresslevel=compresslevel)

    else:
        f = gzip.open(destination, "wb", compresslevel=compresslevel)
    with f:
        for chunk in chunks:
            #print("write gzip", file=sys.stderr)
            f.write(chunk)
    

def read_main(filename, run_raw, run_gzio, run_gzip):
    import time

    expected_size = None
    for run_flag, run_name, run_func in (
            (run_raw, "raw", read_raw_main),
            (run_gzio, "gzio", read_gzio_main),
            (run_gzip, "gzip", read_gzip_main),
            ):
        if not run_flag:
            continue
        t1 = time.time()
        size = run_func(filename)
        t2 = time.time()
        print("read %s: %.2f" % (run_name, t2-t1))
        if expected_size is None:
            expected_size = size
        else:
            assert size == expected_size, (size, expected_size)

def write_main():
    import time
    import argparse
    
    parser = argparse.ArgumentParser()
    parser.add_argument("--level", "-l", type=int, default=9)
    parser.add_argument("--input", "-i", default="/Users/dalke/databases/pubchem_10000.fps")
    parser.add_argument("-n", type=int, default=10)
    parser.add_argument("--output", "-o", default=None)
    parser.add_argument("--method", "-m", choices=("api", "gzip"), default="api")

    args = parser.parse_args()
    
    chunks = get_write_chunks(args.input, args.n)
    call_args = (args.output, _get_compresslevel(args.level), chunks)
    if args.method == "api":
        main = write_api_main
    elif args.method == "gzip":
        main = write_gzip_main
    else:
        raise AssertionError(args.method)
    t1 = time.time()
    main(*call_args)
    t2 = time.time()
    print(main.__name__, t2-t1, file=sys.stderr)
    import subprocess
    if args.output is not None:
        print("md5 input:", subprocess.check_output(["md5sum", args.input]), file=sys.stderr)
        p1 = subprocess.Popen(["gzcat", args.output], stdout=subprocess.PIPE)
        p2 = subprocess.Popen(["md5sum"], stdin=p1.stdout, stdout=subprocess.PIPE)
        output = p2.stdout.read()
        p2.stdout.close()
        p2.wait()
        p1.wait()
        print("md5 output:", output, file=sys.stderr)
        
def main():
    import argparse
    parser = argparse.ArgumentParser("test chemfp's own gzip reader/writer")
    ## parser.add_argument("--write", action="store_true",
    ##                         help="do write timings instead of read timings")
    parser.add_argument("--all", action="store_true",
                            help="same as specifying --raw --gzio --gzip")
    parser.add_argument("--raw", action="store_true",
                            help="time the raw zlib interface")
    parser.add_argument("--gzio", action="store_true",
                            help="time the chemfp.gzio interface")
    parser.add_argument("--gzip", action="store_true",
                            help="time Python's gzip interface")
    parser.add_argument("filename", nargs="?", default=None)

    args = parser.parse_args()

    if args.all:
        args.raw = args.gzio = args.gzip = True
    
    n = bool(args.raw) + bool(args.gzio) + bool(args.gzip)
    
    if 0 and args.write:
        main_func = write_main
        where = "writing to stdout"
    else:
        main_func = read_main
        where = "reading from stdin"

    if args.filename is None:
        if n == 0:
            args.gzio = True
        elif n > 2:
            parser.error("Can only specify one of --raw, --gzio, or --gzip when " + where)
    else:
        if n == 0:
            args.raw = args.gzio = args.gzip = True
            
    main_func(args.filename, args.raw, args.gzio, args.gzip)

if __name__ == "__main__":
    if 1:
        main()
    else:
        write_main()
