# Copyright (c) 2010-2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import absolute_import, print_function

# OpenBabel fingerprints are stored as vector<unsigned int>.  On all
# the machines I use, ints have 32 bits.

# OpenBabel bit lengths must be at least sizeof(int)*8 bits long and
# must be a factor of two. I have no idea why this is required.

# OpenBabel supports new fingerprints through a plugin system.  I got
# it working thanks to Noel O'Boyle's excellent work with Cinfony. I
# then found out that the OB API doesn't have any way to get the
# number of bits in the fingerprint. The size is rounded up to the
# next power of two, so FP4 (307 bits) needs 512 bits (16 ints)
# instead of 320 bits (10 ints). That means I can't even get close to
# guessing the bitsize.

# In the end, I hard-coded the supported fingerprints into the system.

import struct

from . import types
from . import openbabel_toolkit
from . import _openbabel_toolkit

import openbabel
if getattr(openbabel, "__version__", "2.0")[:2] == "2.":
    import openbabel as ob
else:
    import openbabel.openbabel as ob
del openbabel

SOFTWARE = openbabel_toolkit.software
from . import SOFTWARE as CHEMFP_SOFTWARE

# This is a "standard" size according to the struct module
# documentation, so the following is an excess of caution
if struct.calcsize("<I") != 4:
    raise AssertionError("The chemfp.ob module assumes 32 bit integers")


############

# I could have written a more general function which created these but
# there's only a few fingerprints lengths to worry about.

# This needs 128 bytes, for 1024 bits
# vectorUnsignedInt will contain 32 32-bit words = 1024 bits

_ob_get_fingerprint = {}
ECFP_AVAILABLE = False # Meaning, "available in Open Babel"
HAS_ECFP = False  # Meaning, supported by chemfp (only for Open Babel 3.0 and later)
def _init():
    # 
    ob.OBConversion()
    for name in ("FP2", "FP3", "FP4", "MACCS", "ECFP0", "ECFP2", "ECFP4", "ECFP6", "ECFP8", "ECFP10"):
        ob_fingerprinter = ob.OBFingerprint.FindFingerprint(name)
        if ob_fingerprinter is None:
            _ob_get_fingerprint[name] = (None, None)
        else:
            _ob_get_fingerprint[name] = (ob_fingerprinter, ob_fingerprinter.GetFingerprint)

    global ECFP_AVAILABLE, HAS_ECFP
    if _ob_get_fingerprint["ECFP2"][0] is not None:
        ECFP_AVAILABLE = True
        if ob.OBReleaseVersion()[:2] != "2.":
            HAS_ECFP = True
            
    if _ob_get_fingerprint["FP2"][0] is None:
        raise ImportError("Unable to load OpenBabel FP2 fingerprinter. Check $BABEL_LIBDIR")
    n = _ob_get_fingerprint["FP2"][0].Getbitsperint()
    if n != 32:
        raise AssertionError(
            "The chemfp.openbabel_types module assumes Open Babel fingerprints have 32 bit integers")
            
_init()

### TODO: Why do I have calc_FP2, calc_FP3, and calc_FP4?
# They were part of the original API. The calc_MACCS() is used for
# fingerprint detection. But they aren't otherwise used.

_pack_1024 = struct.Struct("<" + "I"*32).pack
## def calc_FP2(mol, fp=None,
##              get_fingerprint=_ob_get_fingerprint["FP2"][1],
##              _pack_1024 = _pack_1024):
##     if fp is None:
##         fp = ob.vectorUnsignedInt()
##     get_fingerprint(mol, fp)
##     return _pack_1024(*fp)

## # This needs 7 bytes, for 56 bits.
## # vectorUnsignedInt will contain 2 32-bit words = 64 bits
_pack_64 = struct.Struct("<II").pack
## def calc_FP3(mol, fp=None,
##              get_fingerprint=_ob_get_fingerprint["FP3"][1],
##              _pack_64 = _pack_64):
##     if fp is None:
##         fp = ob.vectorUnsignedInt()
##     get_fingerprint(mol, fp)
##     return _pack_64(*fp)[:7]

## # This needs 39 bytes, for 312 bits
## # vectorUnsignedInt will contain 16 32-bit words = 512 bits
_pack_512 = struct.Struct("<" + "I"*16).pack
## def calc_FP4(mol, fp=None,
##              get_fingerprint=_ob_get_fingerprint["FP4"][1],
##              _pack_512 = _pack_512):
##     if fp is None:
##         fp = ob.vectorUnsignedInt()
##     get_fingerprint(mol, fp)
##     return _pack_512(*fp)[:39]

# This needs 21 bytes, for 166 bits
# vectorUnsignedInt will contain 8 32-bit words = 256 bits
# (Remember, although 6 words * 32-bits/word = 192, the OpenBabel
# fingerprint size must be a power of 2, and the closest is 8*32.)
_pack_256 = struct.Struct("<" + "I"*8).pack
def calc_MACCS(mol, fp=None,
               get_fingerprint=_ob_get_fingerprint["MACCS"][1],
               _pack_256 = _pack_256):
    if fp is None:
        fp = ob.vectorUnsignedInt()
    get_fingerprint(mol, fp)
    return _pack_256(*fp)[:21]

# TODO: calc_ECFP<n> versions. If they are really needed.
               


# OpenBabel version up to 2.3.0 contained errors in the
# translation of the MACCS patterns from RDKit.
# Post-2.3.0 fixed in version control.
# MACCS might also be missing if BABEL_DATADIR doesn't exist.
HAS_MACCS = False
MACCS_VERSION = 0

def _check_for_maccs():
    global HAS_MACCS, MACCS_VERSION
    if _ob_get_fingerprint["MACCS"] == (None, None):
        if _openbabel_toolkit._ob_version.startswith("2.2."):
            return
        # MACCS should be here. Report the most likely reason
        import os, warnings
        if "BABEL_DATADIR" not in os.environ:
            warnings.warn("MACCS fingerprint missing; perhaps due to missing BABEL_DATADIR?")
        else:
            warnings.warn("MACCS fingerprint missing; perhaps due to BABEL_DATADIR?")
        return

    HAS_MACCS = 1

    # OpenBabel 2.3.0 released the MACCS keys but with a bug in the SMARTS.
    # While they are valid substructure keys, they are not really MACCS keys.
    # This is a run-time detection to figure out which version was installed
    obconversion = ob.OBConversion()
    obconversion.SetInFormat("smi")
    obmol = ob.OBMol()
    obconversion.ReadString(obmol, "CC1=CC(=NN1CC(=O)NNC(=O)C=CC2=C(C=CC=C2Cl)F)C")
    fp = calc_MACCS(obmol)
    if fp == b"\x80\x04\x00\x00\x00\x02\x08\x00\x19\xc4@\xea\xcdl\x98\x0b\xae\xa1x\xef\x1b":
        MACCS_VERSION = 1
    elif fp == b"\x00\x00\x00\x00\x00\x02\x08\x00\x19\xc4D\xea\xcdl\x98\x0b\xae\xa1x\xef\x1f":
        MACCS_VERSION = 2
    else:
        raise AssertionError("Unknown MACCS fingerprint version: %r" % (fp,))

    # There was a fix in 20 October 2014 to support key 44.  I suggested
    # the change. However, I forgot to update chemfp to distinguish
    # between those two fingerprint versions.  Hopefully no one is using
    # that version, but if they are, here's a warning.
    if MACCS_VERSION == 2:
        obmol = ob.OBMol()
        obconversion.ReadString(obmol, "[He]")
        fp = calc_MACCS(obmol)
        if fp != b"\x00\x00\x00\x00\x00\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00":
            import sys
            sys.stderr.write("WARNING: chemfp will use an invalid version for the Open Babel MACCS keys.\n")

_check_for_maccs()

####

class OpenBabelBaseFingerprintType(types.FingerprintType):
    software = SOFTWARE + " " + CHEMFP_SOFTWARE
    toolkit = openbabel_toolkit

    # The Open Babel fingerprinters are not re-entrant so I need
    # to make a new fingerprinter for each make_fingerprinter()

class OpenBabelFP2FingerprintType_v1(types.NoFingerprintParametersMixin, OpenBabelBaseFingerprintType):
    """OpenBabel FP2 fingerprint based on path enumeration

    See http://openbabel.org/wiki/FP2

    This is a Daylight-like path enumeration fingerprint with 1021 bits.

    The OpenBabel-FP2/1 :class:`.FingerprintType` has no parameters.
    """
    name = "OpenBabel-FP2/1"
    num_bits = 1021
    is_available = True

    def make_fingerprinter(self):
        get_fingerprint = _ob_get_fingerprint["FP2"][1]
        pack_1024 = _pack_1024
        def openbabel_fp2_fingerprinter(mol):
            fp = ob.vectorUnsignedInt()
            get_fingerprint(mol, fp)
            return pack_1024(*fp)
        return openbabel_fp2_fingerprinter

class OpenBabelFP3FingerprintType_v1(types.NoFingerprintParametersMixin, OpenBabelBaseFingerprintType):
    """OpenBabel FP3 fingerprint

    See http://openbabel.org/wiki/FP3

    55 bit fingerprints based on a set of SMARTS patterns defining functional groups.

    The OpenBabel-FP3/1 :class:`.FingerprintType` has no parameters.
    """
    name = "OpenBabel-FP3/1"
    num_bits = 55
    is_available = True
    
    def make_fingerprinter(self):
        get_fingerprint = _ob_get_fingerprint["FP3"][1]
        pack_64 = _pack_64
        def openbabel_fp3_fingerprinter(mol):
            fp = ob.vectorUnsignedInt()
            get_fingerprint(mol, fp)
            return pack_64(*fp)[:7]
        return openbabel_fp3_fingerprinter


class OpenBabelFP4FingerprintType_v1(types.NoFingerprintParametersMixin, OpenBabelBaseFingerprintType):
    """OpenBabel FP4 fingerprint

    http://openbabel.org/wiki/FP4

    307 bit fingerprints based on a set of SMARTS patterns defining functional groups.

    The OpenBabel-FP4/1 :class:`.FingerprintType` has no parameters.
    """
    name = "OpenBabel-FP4/1"
    num_bits = 307
    is_available = True
    
    def make_fingerprinter(self):
        get_fingerprint = _ob_get_fingerprint["FP4"][1]
        pack_512 = _pack_512
        def openbabel_fp4_fingerprinter(mol):
            fp = ob.vectorUnsignedInt()
            get_fingerprint(mol, fp)
            return pack_512(*fp)[:39]
        return openbabel_fp4_fingerprinter


class _OpenBabelMACCSFingerprintType(types.NoFingerprintParametersMixin, OpenBabelBaseFingerprintType):
    num_bits = 166

    def make_fingerprinter(self):
        get_fingerprint = _ob_get_fingerprint["MACCS"][1]
        pack_256 = _pack_256
        def openbabel_maccs_fingerprinter(mol):
            fp = ob.vectorUnsignedInt()
            get_fingerprint(mol, fp)
            return pack_256(*fp)[:21]
        return openbabel_maccs_fingerprinter

class OpenBabelMACCSFingerprintType_v1(_OpenBabelMACCSFingerprintType):
    """Open Babel's implementation of the 166 MACCS keys

    WARNING: This implementation contains serious bugs! All of
    the ring sizes are wrong.

    See http://openbabel.org/wiki/Tutorial:Fingerprints and
    https://github.com/openbabel/openbabel/blob/master/data/MACCS.txt .

    The OpenBabel-MACCS/1 :class:`.FingerprintType` has no parameters.

    Note: this version is only available in older (pre-2012) versions of Open Babel.
    """
    name = "OpenBabel-MACCS/1"
    is_available = (MACCS_VERSION == 1)
    if not is_available:
        unavailable_message = "This version of Open Babel does not support OpenBabel-MACCS/1 fingerprints"

class OpenBabelMACCSFingerprintType_v2(_OpenBabelMACCSFingerprintType):
    """Open Babel's implementation of the 166 MACCS keys

    See http://openbabel.org/wiki/Tutorial:Fingerprints and
    https://github.com/openbabel/openbabel/blob/master/data/MACCS.txt .

    Note: Open Babel added support for key 44 on 20 October 2014.  This
    should have been version 3. However, I didn't notice until 1 May
    2017 that there was no chemfp test for it. Since everyone has been
    using it as v2, and very few people used the older version, I
    won't change the version number.

    The OpenBabel-MACCS/2 :class:`.FingerprintType` has no parameters.
    """
    name = "OpenBabel-MACCS/2"
    is_available = (MACCS_VERSION == 2)
    if not is_available:
        unavailable_message = "This version of Open Babel does not support OpenBabel-MACCS/2 fingerprints"

######### ECFP-related code

# Make an ECFP fingerprint function specific to the given fingerprint parameters
ECFP_NUM_BITS = 4096

ecfp_arguments = types.FingerprintArguments()
_add = ecfp_arguments.add_argument

_add("nBits", decoder=types.power_of_2_at_least_32, metavar="INT", default=ECFP_NUM_BITS,
     help = "number of bits in the fingerprint")

# Create the pack function on-demand
class PackTable(dict):
    def __missing__(self, nBits):
        if ((nBits-1) & nBits): # ensure it's a power of 2
            raise ValueError("nBits must be a power of 2: %r" % (nBits,))
        if nBits < 32:
            raise ValueError("nBits is too small: %r" % (nBits,))
        pack = struct.Struct("<" + "I"*(nBits // 32)).pack
        self[nBits] = pack
        return pack

_pack_table = PackTable()
_pack_4096 = _pack_table[4096]

class OpenBabelECFPBaseFingerprintType(OpenBabelBaseFingerprintType):
    _fingerprinter = _get_fingerprint = None
    
    def __init__(self, fingerprint_kwargs):
        super(OpenBabelECFPBaseFingerprintType, self).__init__(fingerprint_kwargs)
        self.num_bits = fingerprint_kwargs["nBits"]

    def _make_fingerprinter(self, nBits=ECFP_NUM_BITS):
        get_fingerprint = self._get_fingerprint
        if nBits == 4096:
            pack = _pack_4096
            def openbabel_ecfp_fingerprinter(mol):
                fp = ob.vectorUnsignedInt()
                get_fingerprint(mol, fp)
                return pack(*fp)

        else:
            fingerprinter = self._fingerprinter
            pack = _pack_table[nBits]
            def openbabel_ecfp_fingerprinter(mol):
                fp = ob.vectorUnsignedInt()
                get_fingerprint(mol, fp)
                fingerprinter.Fold(fp, nBits)
                return pack(*fp)
        
        return openbabel_ecfp_fingerprinter

    _arguments = ecfp_arguments
    _format_string = "nBits={nBits}"

    if not HAS_ECFP:
        if ECFP_AVAILABLE:
            unavailable_message = "chemfp does not support pre-Open Babel 3.0 ECFP fingerprints"
        else:
            unavailable_message = "This version of Open Babel does not implement the ECFP fingerprints"
    
    
class OpenBabelECFP0FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP0 fingerprint

    This is a circular fingerprint of diameter 0.
    
    The OpenBabel-ECFP0/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP0/1"
    is_available = HAS_ECFP
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP0"]

class OpenBabelECFP2FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP2 fingerprint

    This is a circular fingerprint of diameter 2.
    
    The OpenBabel-ECFP2/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP2/1"
    is_available = HAS_ECFP
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP2"]

class OpenBabelECFP4FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP4 fingerprint

    This is a circular fingerprint of diameter 4.
    
    The OpenBabel-ECFP4/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP4/1"
    is_available = HAS_ECFP
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP4"]

class OpenBabelECFP6FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP6 fingerprint

    This is a circular fingerprint of diameter 6.
    
    The OpenBabel-ECFP6/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP6/1"
    is_available = HAS_ECFP
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP6"]

class OpenBabelECFP8FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP8 fingerprint

    This is a circular fingerprint of diameter 8.
    
    The OpenBabel-ECFP8/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP8/1"
    is_available = HAS_ECFP
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP8"]

class OpenBabelECFP10FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP10 fingerprint

    This is a circular fingerprint of diameter 10.
    
    The OpenBabel-ECFP10/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP10/1"
    is_available = HAS_ECFP
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP10"]
                    
    
#########
        
def resolve_chemfp_fingerprint_types(registry):
    for fingerprint_type in (OpenBabelFP2FingerprintType_v1,
                             OpenBabelFP3FingerprintType_v1,
                             OpenBabelFP3FingerprintType_v1,
                             OpenBabelFP4FingerprintType_v1,
                             OpenBabelMACCSFingerprintType_v1,
                             OpenBabelMACCSFingerprintType_v2,
                             OpenBabelECFP0FingerprintType_v1,
                             OpenBabelECFP2FingerprintType_v1,
                             OpenBabelECFP4FingerprintType_v1,
                             OpenBabelECFP6FingerprintType_v1,
                             OpenBabelECFP8FingerprintType_v1,
                             OpenBabelECFP10FingerprintType_v1,
                             ):
        if fingerprint_type.is_available:
            registry.register_fingerprint_type(fingerprint_type)
        else:
            registry.register_unavailable(fingerprint_type.name, fingerprint_type.unavailable_message)
